import os
import textwrap
import pytest

from oduflow.git_analysis import (
    classify_changes,
    _get_module_name,
    _is_security_path,
    _is_data_path,
    _extract_field_lines,
    _check_field_changes,
)


class TestGetModuleName:
    def test_module_file(self):
        assert _get_module_name("sale/models/sale.py") == "sale"

    def test_root_file(self):
        assert _get_module_name("setup.py") is None

    def test_nested(self):
        assert _get_module_name("crm/views/crm_lead.xml") == "crm"


class TestIsSecurityPath:
    def test_security_dir(self):
        assert _is_security_path("sale/security/ir.model.access.csv") is True

    def test_not_security(self):
        assert _is_security_path("sale/views/sale_order.xml") is False

    def test_security_xml(self):
        assert _is_security_path("crm/security/crm_security.xml") is True


class TestIsDataPath:
    def test_data_dir(self):
        assert _is_data_path("sale/data/sale_data.xml") is True

    def test_not_data(self):
        assert _is_data_path("sale/views/sale_order.xml") is False

    def test_nested_data(self):
        assert _is_data_path("addons_ee/connect_elevenlabs/data/tools.xml") is True


class TestClassifyChanges:
    def test_empty(self):
        result = classify_changes([], "/tmp")
        assert result["action"] == "none"

    def test_only_xml_not_security(self):
        files = ["sale/views/sale_order.xml", "crm/views/crm_lead.xml"]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "refresh"
        assert result["modules_to_upgrade"] == []

    def test_py_files(self):
        files = ["sale/models/sale.py"]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "restart"

    def test_security_xml(self):
        files = ["sale/security/ir.model.access.csv"]
        result = classify_changes(files, "/tmp")
        # csv is not xml, should be refresh
        assert result["action"] == "refresh"

    def test_security_xml_file(self):
        files = ["sale/security/sale_security.xml"]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "upgrade"
        assert "sale" in result["modules_to_upgrade"]

    def test_data_xml_triggers_upgrade(self):
        files = ["connect_elevenlabs/data/tools.xml"]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "upgrade"
        assert "connect_elevenlabs" in result["modules_to_upgrade"]

    def test_js_only(self):
        files = ["web/static/src/js/app.js"]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "refresh"

    def test_py_plus_xml(self):
        files = ["sale/models/sale.py", "sale/views/sale_order.xml"]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "restart"

    def test_upgrade_beats_restart(self):
        files = [
            "sale/models/sale.py",
            "sale/security/sale_security.xml",
        ]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "upgrade"
        assert "sale" in result["modules_to_upgrade"]

    def test_manifest_version_change(self, tmp_path):
        module_dir = tmp_path / "sale"
        module_dir.mkdir()

        old_manifest = "{'name': 'Sale', 'version': '15.0.1.0.0', 'data': []}"
        new_manifest = "{'name': 'Sale', 'version': '15.0.1.1.0', 'data': []}"
        (module_dir / "__manifest__.py").write_text(new_manifest)

        os.makedirs(tmp_path / ".git", exist_ok=True)

        import subprocess
        from unittest.mock import patch

        with patch("subprocess.run") as mock_run:
            mock_rev = type("Result", (), {"stdout": old_manifest})()
            mock_run.return_value = mock_rev

            files = ["sale/__manifest__.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "upgrade"
            assert "sale" in result["modules_to_upgrade"]

    def test_manifest_data_list_change(self, tmp_path):
        module_dir = tmp_path / "crm"
        module_dir.mkdir()

        new_manifest = "{'name': 'CRM', 'version': '15.0.1.0.0', 'data': ['views/crm.xml', 'views/new.xml']}"
        (module_dir / "__manifest__.py").write_text(new_manifest)

        import subprocess
        from unittest.mock import patch

        old_manifest = "{'name': 'CRM', 'version': '15.0.1.0.0', 'data': ['views/crm.xml']}"
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = type("Result", (), {"stdout": old_manifest})()

            files = ["crm/__manifest__.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "upgrade"
            assert "crm" in result["modules_to_upgrade"]

    def test_manifest_no_significant_change(self, tmp_path):
        module_dir = tmp_path / "sale"
        module_dir.mkdir()

        new_manifest = "{'name': 'Sale Updated', 'version': '15.0.1.0.0', 'data': []}"
        (module_dir / "__manifest__.py").write_text(new_manifest)

        import subprocess
        from unittest.mock import patch

        old_manifest = "{'name': 'Sale', 'version': '15.0.1.0.0', 'data': []}"
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = type("Result", (), {"stdout": old_manifest})()

            files = ["sale/__manifest__.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "refresh"

    def test_multiple_modules_upgrade(self):
        files = [
            "sale/security/sale_security.xml",
            "crm/security/crm_security.xml",
        ]
        result = classify_changes(files, "/tmp")
        assert result["action"] == "upgrade"
        assert sorted(result["modules_to_upgrade"]) == ["crm", "sale"]

    def test_nested_addon_dir(self, tmp_path):
        module_dir = tmp_path / "addons_veles" / "customer_code"
        module_dir.mkdir(parents=True)
        (module_dir / "__manifest__.py").write_text(
            "{'name': 'Customer Code', 'version': '17.0.1.0.0', 'data': []}"
        )

        from unittest.mock import patch

        old_manifest = "{'name': 'Customer Code', 'version': '17.0.0.0.0', 'data': []}"
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = type("Result", (), {"stdout": old_manifest})()

            files = ["addons_veles/customer_code/__manifest__.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "upgrade"
            assert result["modules_to_upgrade"] == ["customer_code"]

    def test_new_module_install(self, tmp_path):
        module_dir = tmp_path / "addons_veles" / "customer_code"
        module_dir.mkdir(parents=True)
        (module_dir / "__manifest__.py").write_text(
            "{'name': 'Customer Code', 'version': '17.0.1.0.0', 'data': []}"
        )
        models_dir = module_dir / "models"
        models_dir.mkdir()
        (models_dir / "res_partner.py").write_text(
            "from odoo import fields, models\n\n"
            "class ResPartner(models.Model):\n"
            "    customer_code = fields.Char(string='Customer Code')\n"
        )

        import subprocess
        from unittest.mock import patch

        def mock_subprocess_run(cmd, **kwargs):
            raise subprocess.CalledProcessError(128, cmd)

        with patch("subprocess.run", side_effect=mock_subprocess_run):
            files = [
                "addons_veles/customer_code/__manifest__.py",
                "addons_veles/customer_code/__init__.py",
                "addons_veles/customer_code/models/__init__.py",
                "addons_veles/customer_code/models/res_partner.py",
            ]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "install"
            assert result["modules_to_install"] == ["customer_code"]
            assert result["modules_to_upgrade"] == []

    def test_nested_addon_dir_py_file(self, tmp_path):
        module_dir = tmp_path / "addons_ee" / "sale_ext"
        module_dir.mkdir(parents=True)
        (module_dir / "__manifest__.py").write_text(
            "{'name': 'Sale Ext', 'version': '17.0.1.0.0'}"
        )

        files = ["addons_ee/sale_ext/security/ir.model.access.xml"]
        result = classify_changes(files, str(tmp_path))
        assert result["action"] == "upgrade"
        assert result["modules_to_upgrade"] == ["sale_ext"]

    def test_py_field_added_triggers_upgrade(self, tmp_path):
        module_dir = tmp_path / "sale" / "models"
        module_dir.mkdir(parents=True)
        (tmp_path / "sale" / "__manifest__.py").write_text(
            "{'name': 'Sale', 'version': '17.0.1.0.0'}"
        )
        (module_dir / "sale.py").write_text(textwrap.dedent("""\
            from odoo import fields, models

            class SaleOrder(models.Model):
                name = fields.Char(string='Name')
                customer_code = fields.Char(string='Customer Code')
        """))

        from unittest.mock import patch

        old_source = textwrap.dedent("""\
            from odoo import fields, models

            class SaleOrder(models.Model):
                name = fields.Char(string='Name')
        """)
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = type("Result", (), {"stdout": old_source})()

            files = ["sale/models/sale.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "upgrade"
            assert "sale" in result["modules_to_upgrade"]

    def test_py_field_removed_triggers_upgrade(self, tmp_path):
        module_dir = tmp_path / "crm" / "models"
        module_dir.mkdir(parents=True)
        (tmp_path / "crm" / "__manifest__.py").write_text(
            "{'name': 'CRM', 'version': '17.0.1.0.0'}"
        )
        (module_dir / "lead.py").write_text(textwrap.dedent("""\
            from odoo import fields, models

            class CrmLead(models.Model):
                name = fields.Char(string='Name')
        """))

        from unittest.mock import patch

        old_source = textwrap.dedent("""\
            from odoo import fields, models

            class CrmLead(models.Model):
                name = fields.Char(string='Name')
                priority = fields.Selection(string='Priority')
        """)
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = type("Result", (), {"stdout": old_source})()

            files = ["crm/models/lead.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "upgrade"
            assert "crm" in result["modules_to_upgrade"]

    def test_py_field_param_changed_triggers_upgrade(self, tmp_path):
        module_dir = tmp_path / "sale" / "models"
        module_dir.mkdir(parents=True)
        (tmp_path / "sale" / "__manifest__.py").write_text(
            "{'name': 'Sale', 'version': '17.0.1.0.0'}"
        )
        (module_dir / "sale.py").write_text(textwrap.dedent("""\
            from odoo import fields, models

            class SaleOrder(models.Model):
                name = fields.Char(string='Name', required=True)
        """))

        from unittest.mock import patch

        old_source = textwrap.dedent("""\
            from odoo import fields, models

            class SaleOrder(models.Model):
                name = fields.Char(string='Name')
        """)
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = type("Result", (), {"stdout": old_source})()

            files = ["sale/models/sale.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "upgrade"
            assert "sale" in result["modules_to_upgrade"]

    def test_py_no_field_change_stays_restart(self, tmp_path):
        module_dir = tmp_path / "sale" / "models"
        module_dir.mkdir(parents=True)
        (tmp_path / "sale" / "__manifest__.py").write_text(
            "{'name': 'Sale', 'version': '17.0.1.0.0'}"
        )
        (module_dir / "sale.py").write_text(textwrap.dedent("""\
            from odoo import fields, models

            class SaleOrder(models.Model):
                name = fields.Char(string='Name')

                def action_confirm(self):
                    return True
        """))

        from unittest.mock import patch

        old_source = textwrap.dedent("""\
            from odoo import fields, models

            class SaleOrder(models.Model):
                name = fields.Char(string='Name')

                def action_confirm(self):
                    pass
        """)
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = type("Result", (), {"stdout": old_source})()

            files = ["sale/models/sale.py"]
            result = classify_changes(files, str(tmp_path))
            assert result["action"] == "restart"
            assert result["modules_to_upgrade"] == []


class TestExtractFieldLines:
    def test_basic(self):
        source = "    name = fields.Char(string='Name')\n"
        assert _extract_field_lines(source) == {"name = fields.Char(string='Name')"}

    def test_no_spaces(self):
        source = "customer_code=fields.Char(string='Code')\n"
        assert _extract_field_lines(source) == {"customer_code=fields.Char(string='Code')"}

    def test_many2one(self):
        source = "    partner_id = fields.Many2one('res.partner')\n"
        assert _extract_field_lines(source) == {"partner_id = fields.Many2one('res.partner')"}

    def test_no_fields(self):
        source = "def create(self, vals):\n    return super().create(vals)\n"
        assert _extract_field_lines(source) == set()



import os
import re
from urllib.parse import urlparse, urlunparse


def slugify_branch(branch_name: str) -> str:
    slug = branch_name.replace("/", "-")
    slug = re.sub(r"[^a-zA-Z0-9_-]", "", slug)
    slug = slug.lower()
    return slug[:63]


def get_db_name(branch_name: str, instance_id: str = "1") -> str:
    return f"oduflow_{instance_id}_{slugify_branch(branch_name)}"


def get_resource_name(branch_name: str, resource_type: str, prefix: str = "oduflow-") -> str:
    return f"{prefix}{branch_name.replace('/', '-')}-{resource_type}"


def get_workspace_path(branch_name: str, workspaces_dir: str) -> str:
    return os.path.join(workspaces_dir, branch_name.replace("/", "-"))


def get_repo_path(branch_name: str, workspaces_dir: str) -> str:
    return os.path.join(get_workspace_path(branch_name, workspaces_dir), "repo")


def get_env_hostname(branch_name: str, base_domain: str) -> str:
    slug = slugify_branch(branch_name).replace("_", "-")
    slug = re.sub(r"-+", "-", slug).strip("-")
    return f"{slug}.{base_domain}"


def get_template_db_name(template_name: str = "default", instance_id: str = "1") -> str:
    slug = template_name.replace("/", "-")
    return f"odoo_ref_{instance_id}_{slug}"


def get_filestore_paths(branch_name: str, workspaces_dir: str) -> dict[str, str]:
    base = get_workspace_path(branch_name, workspaces_dir)
    return {
        "upper": os.path.join(base, "filestore_upper"),
        "work": os.path.join(base, "filestore_work"),
        "merged": os.path.join(base, "filestore"),
    }


def sanitize_repo_url(url: str) -> str:
    if not url:
        return url
    try:
        parsed = urlparse(url)
        if parsed.username or parsed.password:
            clean = parsed._replace(netloc=parsed.hostname or "")
            return urlunparse(clean)
    except Exception:
        pass
    return url



"""Manage service preset configurations (save / restore / list / delete).

Presets are persisted as a single JSON file at ``{settings.home}/service_presets.json``.
"""

import json
import logging
import os

from oduflow.errors import NotFoundError
from oduflow.settings import Settings

logger = logging.getLogger("oduflow")


def _presets_path(settings: Settings) -> str:
    """Return the absolute path to the presets JSON file."""
    return os.path.join(settings.home, "service_presets.json")


def _load_presets(settings: Settings) -> dict:
    """Read and return all presets from disk.

    Returns an empty dict when the file does not exist or contains invalid JSON.
    """
    path = _presets_path(settings)
    if not os.path.exists(path):
        return {}
    try:
        with open(path, "r", encoding="utf-8") as fh:
            return json.load(fh)
    except json.JSONDecodeError:
        logger.warning("Corrupt presets file at %s – returning empty dict", path)
        return {}


def _save_presets(settings: Settings, data: dict) -> None:
    """Persist *data* to the presets JSON file, creating parent dirs if needed."""
    path = _presets_path(settings)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w", encoding="utf-8") as fh:
        json.dump(data, fh, indent=2)


def save_preset(
    settings: Settings,
    name: str,
    image: str,
    port: int,
    hostname: str | None = None,
    env_vars: dict[str, str] | None = None,
) -> dict:
    """Save (or overwrite) a single service preset and return it."""
    short_hostname = hostname or ""
    if short_hostname and settings.base_domain:
        suffix = f".{settings.base_domain}"
        if short_hostname.endswith(suffix):
            short_hostname = short_hostname[: -len(suffix)]
    preset = {
        "image": image,
        "port": port,
        "hostname": short_hostname,
        "env_vars": env_vars if env_vars is not None else {},
    }
    data = _load_presets(settings)
    data[name] = preset
    _save_presets(settings, data)
    return preset


def list_presets(settings: Settings) -> list[dict]:
    """Return all presets as a sorted list of dicts (each includes a ``name`` key)."""
    data = _load_presets(settings)
    return sorted(
        [{"name": name, **preset} for name, preset in data.items()],
        key=lambda p: p["name"],
    )


def get_preset(settings: Settings, name: str) -> dict:
    """Return a single preset dict (with ``name`` key) or raise :class:`NotFoundError`."""
    data = _load_presets(settings)
    if name not in data:
        raise NotFoundError(f"Service preset '{name}' not found")
    return {"name": name, **data[name]}


def delete_preset(settings: Settings, name: str) -> dict:
    """Remove a preset from disk and return ``{"name": name}``.

    Raises :class:`NotFoundError` if the preset does not exist.
    """
    data = _load_presets(settings)
    if name not in data:
        raise NotFoundError(f"Service preset '{name}' not found")
    del data[name]
    _save_presets(settings, data)
    return {"name": name}



import logging
import re

import docker

from oduflow.docker_ops.client import get_client
from oduflow.docker_ops import service_presets
from oduflow.errors import ConflictError, NotFoundError, PrerequisiteNotMetError
from oduflow.settings import Settings

logger = logging.getLogger("oduflow")

_SYSTEM_ENV_KEYS = {
    "PATH", "HOME", "HOSTNAME", "TERM", "LANG", "LC_ALL",
    "GOPATH", "JAVA_HOME", "SHLVL", "PWD", "OLDPWD", "SHELL",
    "USER", "LOGNAME", "MAIL", "EDITOR", "VISUAL", "PAGER",
    "LESS", "LESSOPEN", "LESSCLOSE", "LS_COLORS",
    "XDG_RUNTIME_DIR", "XDG_DATA_DIRS", "XDG_CONFIG_DIRS", "XDG_CACHE_HOME",
    "DISPLAY", "GPG_AGENT_INFO", "SSH_AUTH_SOCK", "SSH_AGENT_PID",
    "DBUS_SESSION_BUS_ADDRESS",
}


def create_service(
    settings: Settings,
    name: str,
    image: str,
    port: int,
    hostname: str | None = None,
    env_vars: dict[str, str] | None = None,
) -> dict[str, str]:
    client = get_client()
    container_name = f"oduflow-svc-{name}"

    # Check that the shared network exists
    try:
        client.networks.get(settings.shared_network)
    except docker.errors.NotFound:
        raise PrerequisiteNotMetError(
            f"Shared network '{settings.shared_network}' not found. "
            "Run init_system first."
        )

    # Check for existing container
    try:
        existing = client.containers.get(container_name)
        if existing.status == "running":
            raise ConflictError(
                f"Service '{name}' already exists and is running."
            )
    except docker.errors.NotFound:
        pass

    labels = {
        "oduflow.managed": "true",
        "oduflow.instance": settings.instance_id,
        "oduflow.service": name,
    }

    run_kwargs: dict = {
        "image": image,
        "name": container_name,
        "detach": True,
        "network": settings.shared_network,
        "labels": labels,
        "restart_policy": {"Name": "unless-stopped"},
    }

    if settings.routing_mode == "traefik":
        if not hostname:
            hostname = f"{name}.{settings.base_domain}"
        elif "." not in hostname:
            hostname = f"{hostname}.{settings.base_domain}"
        labels["traefik.enable"] = "true"
        labels[f"traefik.http.routers.{container_name}.rule"] = f"Host(`{hostname}`)"
        labels[f"traefik.http.routers.{container_name}.entrypoints"] = "websecure"
        labels[f"traefik.http.routers.{container_name}.tls.certresolver"] = "le"
        labels[f"traefik.http.services.{container_name}.loadbalancer.server.port"] = str(port)
        url = f"https://{hostname}"
    else:
        run_kwargs["ports"] = {f"{port}/tcp": port}
        url = f"http://{settings.external_host}:{port}"

    if env_vars:
        run_kwargs["environment"] = env_vars

    client.containers.run(**run_kwargs)
    logger.info("Created service container %s from image %s", container_name, image)

    # Auto-save preset for future restore
    try:
        service_presets.save_preset(
            settings, name, image, port,
            hostname=hostname,
            env_vars=env_vars,
        )
    except Exception:
        logger.warning("Failed to save service preset for %s", name, exc_info=True)

    return {
        "name": name,
        "container_name": container_name,
        "image": image,
        "url": url,
    }


def delete_service(settings: Settings, name: str) -> dict[str, str]:
    client = get_client()
    container_name = f"oduflow-svc-{name}"

    try:
        container = client.containers.get(container_name)
    except docker.errors.NotFound:
        raise NotFoundError(f"Service '{name}' not found")

    container.stop()
    container.remove(v=True)
    logger.info("Deleted service container %s", container_name)

    return {
        "name": name,
        "container_name": container_name,
    }


def list_services(settings: Settings) -> list[dict]:
    client = get_client()
    containers = client.containers.list(
        all=True,
        filters={
            "label": [
                f"{settings.managed_label}=true",
                f"{settings.instance_label}={settings.instance_id}"
            ]
        },
    )

    result = []
    for container in containers:
        svc_name = container.labels.get("oduflow.service")
        if not svc_name:
            continue

        container_name = container.name
        image = container.image.tags[0] if container.image.tags else "unknown"
        status = container.status

        # Parse environment variables, filtering out system keys
        raw_env = container.attrs.get("Config", {}).get("Env", [])
        env_vars = {}
        for entry in raw_env:
            if "=" in entry:
                key, value = entry.split("=", 1)
                if key not in _SYSTEM_ENV_KEYS:
                    env_vars[key] = value

        # Extract port and URL
        port_num = None
        url = None

        if settings.routing_mode == "traefik":
            rule_label = f"traefik.http.routers.oduflow-svc-{svc_name}.rule"
            port_label = f"traefik.http.services.oduflow-svc-{svc_name}.loadbalancer.server.port"

            rule_value = container.labels.get(rule_label, "")
            match = re.search(r"Host\(`([^`]+)`\)", rule_value)
            if match:
                traefik_hostname = match.group(1)
                url = f"https://{traefik_hostname}"

            label_port = container.labels.get(port_label)
            if label_port:
                port_num = int(label_port)
        else:
            ports_dict = container.attrs.get("NetworkSettings", {}).get("Ports", {})
            if ports_dict:
                for port_key, mappings in ports_dict.items():
                    # Extract port number from key like "6379/tcp"
                    port_match = re.match(r"(\d+)/", port_key)
                    if port_match:
                        port_num = int(port_match.group(1))
                    if mappings:
                        for mapping in mappings:
                            host_port = mapping.get("HostPort")
                            if host_port:
                                url = f"http://{settings.external_host}:{host_port}"
                                break
                    break  # only process first port entry

        result.append({
            "name": svc_name,
            "container_name": container_name,
            "image": image,
            "status": status,
            "port": port_num,
            "url": url,
            "env_vars": env_vars,
        })

    return result


def update_service(settings: Settings, name: str) -> dict[str, str]:
    """Pull the latest image for a service and re-create it with the same settings."""
    client = get_client()
    container_name = f"oduflow-svc-{name}"

    try:
        container = client.containers.get(container_name)
    except docker.errors.NotFound:
        raise NotFoundError(f"Service '{name}' not found")

    # Capture current image name
    old_image = container.image.tags[0] if container.image.tags else None
    if not old_image:
        # Fall back to Config.Image (may be a digest reference)
        old_image = container.attrs.get("Config", {}).get("Image")
    if not old_image:
        raise NotFoundError(
            f"Cannot determine image for service '{name}'. "
            "The container has no image tag or Config.Image."
        )

    # Capture current environment variables (filtering system keys)
    raw_env = container.attrs.get("Config", {}).get("Env", [])
    env_vars: dict[str, str] = {}
    for entry in raw_env:
        if "=" in entry:
            key, value = entry.split("=", 1)
            if key not in _SYSTEM_ENV_KEYS:
                env_vars[key] = value

    # Capture port and hostname depending on routing mode
    port: int | None = None
    hostname: str | None = None

    if settings.routing_mode == "traefik":
        rule_label = f"traefik.http.routers.{container_name}.rule"
        port_label = f"traefik.http.services.{container_name}.loadbalancer.server.port"

        rule_value = container.labels.get(rule_label, "")
        match = re.search(r"Host\(`([^`]+)`\)", rule_value)
        if match:
            hostname = match.group(1)

        label_port = container.labels.get(port_label)
        if label_port:
            port = int(label_port)
    else:
        ports_dict = container.attrs.get("NetworkSettings", {}).get("Ports", {})
        if ports_dict:
            for port_key in ports_dict:
                port_match = re.match(r"(\d+)/", port_key)
                if port_match:
                    port = int(port_match.group(1))
                    break

    if port is None:
        raise NotFoundError(
            f"Cannot determine port for service '{name}'."
        )

    # Capture old image digest
    old_digest = container.image.id  # e.g. sha256:abc...

    # Pull the latest image
    logger.info("Pulling latest image %s for service %s", old_image, name)
    new_image_obj = client.images.pull(old_image)
    new_digest = new_image_obj.id
    image_updated = old_digest != new_digest

    if not image_updated:
        logger.info("Image unchanged for service %s: %s", name, new_digest[:19])
        return {
            "name": name,
            "container_name": container_name,
            "image": old_image,
            "image_updated": False,
            "old_digest": old_digest,
            "new_digest": new_digest,
        }

    logger.info("Image changed for service %s: %s -> %s", name, old_digest[:19], new_digest[:19])

    # Stop and remove the old container
    container.stop()
    container.remove(v=True)
    logger.info("Removed old container %s for update", container_name)

    # Re-create with the same settings
    result = create_service(
        settings,
        name=name,
        image=old_image,
        port=port,
        hostname=hostname,
        env_vars=env_vars or None,
    )

    result["image_updated"] = True
    result["old_digest"] = old_digest
    result["new_digest"] = new_digest

    logger.info("Service %s updated with image %s", name, old_image)
    return result


def get_service_logs(settings: Settings, name: str, lines: int = 100) -> str:
    client = get_client()
    container_name = f"oduflow-svc-{name}"

    try:
        container = client.containers.get(container_name)
    except docker.errors.NotFound:
        raise NotFoundError(f"Service '{name}' not found")

    output = container.logs(tail=lines, timestamps=True)
    return output.decode("utf-8")



import logging
import os
import shutil
import subprocess
import tempfile
from urllib.parse import urlparse

from oduflow.errors import ExternalCommandError, FlowError

logger = logging.getLogger("oduflow")

GIT_ENV = {**os.environ, "GIT_TERMINAL_PROMPT": "0"}


class InvalidRepoURLError(FlowError):
    """Repository URL is invalid or missing credentials."""


class RepoAuthError(FlowError):
    """Repository authentication failed. Call setup_repo_auth first."""


def _parse_authenticated_url(repo_url: str) -> tuple[str, str, str, str]:
    parsed = urlparse(repo_url)
    if parsed.scheme not in ("https", "http"):
        raise InvalidRepoURLError(
            f"URL must use https:// scheme, got: {parsed.scheme}://"
        )
    if not parsed.username or not parsed.password:
        raise InvalidRepoURLError(
            "URL must contain credentials: https://user:PAT@github.com/owner/repo.git"
        )
    clean_url = parsed._replace(netloc=parsed.hostname or "").geturl()
    return clean_url, parsed.hostname or "", parsed.username, parsed.password


def _store_git_credentials(host: str, username: str, password: str) -> None:
    subprocess.run(
        ["git", "config", "--global", "credential.helper", "store"],
        check=True,
        capture_output=True,
        env=GIT_ENV,
    )

    credential_input = (
        f"protocol=https\n"
        f"host={host}\n"
        f"username={username}\n"
        f"password={password}\n"
        f"\n"
    )
    subprocess.run(
        ["git", "credential", "approve"],
        input=credential_input.encode(),
        check=True,
        capture_output=True,
        env=GIT_ENV,
    )
    logger.info("Git credentials stored for host=%s user=%s", host, username)


def setup_repo_auth(repo_url: str) -> dict[str, str]:
    clean_url, host, username, password = _parse_authenticated_url(repo_url)

    _store_git_credentials(host, username, password)

    tmp_dir = tempfile.mkdtemp(prefix="flow-auth-test-")
    try:
        subprocess.run(
            ["git", "clone", "--depth", "1", clean_url, tmp_dir],
            check=True,
            capture_output=True,
            timeout=60,
            env=GIT_ENV,
        )
    except subprocess.CalledProcessError as e:
        error_msg = e.stderr.decode("utf-8") if e.stderr else str(e)
        raise ExternalCommandError(
            "git clone (auth test)", e.returncode,
            f"Credentials saved but test clone failed: {error_msg}",
        )
    except subprocess.TimeoutExpired:
        raise ExternalCommandError(
            "git clone (auth test)", -1,
            "Test clone timed out (60s).",
        )
    finally:
        shutil.rmtree(tmp_dir, ignore_errors=True)

    logger.info("Repo auth verified for %s", clean_url)
    return {"repo_url": clean_url, "host": host, "status": "authenticated"}


def pull_repo(repo_path: str, branch: str) -> list[str]:
    """Pull latest changes and return list of changed file paths (relative to repo root)."""
    old_head = subprocess.run(
        ["git", "-C", repo_path, "rev-parse", "HEAD"],
        check=True,
        capture_output=True,
        text=True,
        env=GIT_ENV,
    ).stdout.strip()

    try:
        subprocess.run(
            ["git", "-C", repo_path, "pull", "--rebase", "origin", branch],
            check=True,
            capture_output=True,
            text=True,
            timeout=60,
            env=GIT_ENV,
        )
    except subprocess.CalledProcessError as e:
        error_msg = e.stderr or str(e)
        raise ExternalCommandError("git pull", e.returncode, error_msg)
    except subprocess.TimeoutExpired:
        raise ExternalCommandError("git pull", -1, "Pull timed out (60s).")

    new_head = subprocess.run(
        ["git", "-C", repo_path, "rev-parse", "HEAD"],
        check=True,
        capture_output=True,
        text=True,
        env=GIT_ENV,
    ).stdout.strip()

    if old_head == new_head:
        return []

    result = subprocess.run(
        ["git", "-C", repo_path, "diff", "--name-only", f"{old_head}..{new_head}"],
        check=True,
        capture_output=True,
        text=True,
        env=GIT_ENV,
    )
    return [f for f in result.stdout.strip().splitlines() if f]


def parse_manifest(manifest_path: str) -> dict:
    """Parse an Odoo __manifest__.py file and return its dict."""
    import ast
    with open(manifest_path, "r") as f:
        return ast.literal_eval(f.read())



import pytest
from oduflow.settings import Settings


class TestSettings:
    def test_defaults(self):
        s = Settings()
        assert s.external_host == "localhost"
        assert s.port_range_start == 50000
        assert s.port_range_end == 50100
        assert s.db_user == "odoo"

    def test_validate_port_range(self):
        s = Settings(port_range_start=50100, port_range_end=50000, workspaces_dir="/tmp")
        with pytest.raises(ValueError, match="Invalid port range"):
            s.validate()

    def test_validate_workspaces_dir(self):
        s = Settings(workspaces_dir="")
        with pytest.raises(ValueError, match="workspaces_dir must be set"):
            s.validate()

    def test_validate_ok(self):
        s = Settings(workspaces_dir="/tmp")
        s.validate()

    def test_from_env(self, monkeypatch):
        monkeypatch.setenv("EXTERNAL_HOST", "10.0.0.1")
        monkeypatch.setenv("PORT_RANGE_START", "60000")
        monkeypatch.setenv("PORT_RANGE_END", "60100")
        s = Settings.from_env()
        assert s.external_host == "10.0.0.1"
        assert s.port_range_start == 60000
        assert s.port_range_end == 60100

    def test_frozen(self):
        s = Settings()
        with pytest.raises(AttributeError):
            s.external_host = "changed"  # type: ignore[misc]


class TestTemplatePaths:
    def test_get_template_dir_default(self):
        s = Settings(home="/srv/data")
        assert s.get_template_dir() == "/srv/data/templates/default"

    def test_get_template_dir_named(self):
        s = Settings(home="/srv/data")
        assert s.get_template_dir("myproject") == "/srv/data/templates/myproject"

    def test_get_template_sql_path(self):
        s = Settings(home="/srv/data")
        assert s.get_template_sql_path("v17") == "/srv/data/templates/v17/dump.pgdump"

    def test_get_template_filestore_path(self):
        s = Settings(home="/srv/data")
        assert s.get_template_filestore_path("v17") == "/srv/data/templates/v17/filestore"

    def test_dump_methods_delegate_to_template(self):
        s = Settings(home="/srv/data")
        assert s.get_dump_sql_path() == s.get_template_sql_path("default")
        assert s.get_dump_filestore_path() == s.get_template_filestore_path("default")

    def test_list_templates_empty(self, tmp_path):
        s = Settings(home=str(tmp_path))
        assert s.list_templates() == []

    def test_list_templates(self, tmp_path):
        templates_dir = tmp_path / "templates"
        (templates_dir / "alpha").mkdir(parents=True)
        (templates_dir / "beta").mkdir(parents=True)
        (templates_dir / "not-a-dir").touch()  # should be ignored
        s = Settings(home=str(tmp_path))
        assert s.list_templates() == ["alpha", "beta"]



class FlowError(Exception):
    """Base error for all Flow operations."""


class BusyError(FlowError):
    """Another operation is already in progress."""


class NotFoundError(FlowError):
    """Environment or resource not found."""


class PrerequisiteNotMetError(FlowError):
    """System not initialized or dependency missing."""


class ProtectedError(FlowError):
    """Environment is protected and cannot be deleted."""


class ConflictError(FlowError):
    """Resource already exists."""


class ExternalCommandError(FlowError):
    """External command (git, psql, docker) failed."""

    def __init__(self, cmd: str, exit_code: int, output: str):
        self.cmd = cmd
        self.exit_code = exit_code
        self.output = output
        super().__init__(f"Command '{cmd}' failed (exit {exit_code}): {output}")



"""
Integration tests driven by fixtures.yaml.

These tests call real MCP tool functions or CLI commands against a live Docker
environment.
Run with:  pytest tests/test_integration.py -m integration -v
Skip with: pytest -m "not integration"

Heavyweight scenarios (init/destroy/create/delete env) are skipped by default.
Run them explicitly:  pytest tests/test_integration.py -m heavyweight -v
"""

from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

import pytest
import yaml

from tool_helpers import call_cli, call_tool

FIXTURES_PATH = Path(__file__).parent / "fixtures.yaml"

MARK_MAP = {
    "heavyweight": pytest.mark.heavyweight,
}


def _load_scenarios():
    with open(FIXTURES_PATH) as f:
        return yaml.safe_load(f)["scenarios"]


SCENARIOS = _load_scenarios()


def _build_params():
    ids = []
    params = []
    for name, scenario in SCENARIOS.items():
        marks = [MARK_MAP[m] for m in scenario.get("marks", []) if m in MARK_MAP]
        ids.append(name)
        params.append(pytest.param(scenario, marks=marks, id=name))
    return params


@pytest.mark.integration
@pytest.mark.parametrize("scenario", _build_params())
def test_scenario(scenario, live_environment):
    if scenario.get("needs_env"):
        settings = live_environment
        with patch("oduflow.server._get_settings", return_value=settings):
            if "cli" in scenario:
                result = call_cli(scenario["cli"], **scenario.get("args", {}))
            else:
                result = call_tool(scenario["tool"], **scenario.get("args", {}))
    else:
        if "cli" in scenario:
            result = call_cli(scenario["cli"], **scenario.get("args", {}))
        else:
            result = call_tool(scenario["tool"], **scenario.get("args", {}))
    assert isinstance(result, str), f"Expected str, got {type(result)}"
    for expected in scenario.get("expect_contains", []):
        assert expected in result, (
            f"Expected '{expected}' in result.\nGot: {result}"
        )



from oduflow.naming import slugify_branch, get_db_name, get_resource_name, get_template_db_name, get_workspace_path, get_repo_path, get_filestore_paths


class TestSlugifyBranch:
    def test_simple(self):
        assert slugify_branch("main") == "main"

    def test_slash(self):
        assert slugify_branch("feature/payments") == "feature-payments"

    def test_complex(self):
        assert slugify_branch("hotfix/CRM-123/fix") == "hotfix-crm-123-fix"

    def test_special_chars(self):
        assert slugify_branch("feat/hello@world!") == "feat-helloworld"

    def test_truncation(self):
        long_name = "a" * 100
        assert len(slugify_branch(long_name)) == 63

    def test_uppercase(self):
        assert slugify_branch("Feature/BRANCH") == "feature-branch"

    def test_underscores_preserved(self):
        assert slugify_branch("fix_bug") == "fix_bug"

    def test_dots_removed(self):
        assert slugify_branch("release/1.2.3") == "release-123"

    def test_empty_string(self):
        assert slugify_branch("") == ""


class TestGetDbName:
    def test_main(self):
        assert get_db_name("main") == "oduflow_1_main"

    def test_feature(self):
        assert get_db_name("feature/payments") == "oduflow_1_feature-payments"

    def test_complex(self):
        assert get_db_name("hotfix/CRM-123/fix") == "oduflow_1_hotfix-crm-123-fix"

    def test_with_instance_id(self):
        assert get_db_name("main", instance_id="2") == "oduflow_2_main"

    def test_feature_with_instance_id(self):
        assert get_db_name("feature/payments", instance_id="3") == "oduflow_3_feature-payments"


class TestGetResourceName:
    def test_odoo(self):
        assert get_resource_name("main", "odoo") == "oduflow-main-odoo"

    def test_slash_branch(self):
        assert get_resource_name("feature/payments", "odoo") == "oduflow-feature-payments-odoo"

    def test_custom_prefix(self):
        assert get_resource_name("main", "odoo", prefix="test-") == "test-main-odoo"


class TestGetWorkspacePath:
    def test_simple(self):
        assert get_workspace_path("main", "/tmp/ws") == "/tmp/ws/main"

    def test_slash(self):
        assert get_workspace_path("feature/payments", "/tmp/ws") == "/tmp/ws/feature-payments"


class TestGetRepoPath:
    def test_simple(self):
        assert get_repo_path("main", "/tmp/ws") == "/tmp/ws/main/repo"

    def test_slash(self):
        assert get_repo_path("feature/payments", "/tmp/ws") == "/tmp/ws/feature-payments/repo"


class TestGetFilestorePaths:
    def test_keys(self):
        paths = get_filestore_paths("main", "/tmp/ws")
        assert set(paths.keys()) == {"upper", "work", "merged"}

    def test_paths(self):
        paths = get_filestore_paths("main", "/tmp/ws")
        assert paths["upper"] == "/tmp/ws/main/filestore_upper"
        assert paths["work"] == "/tmp/ws/main/filestore_work"
        assert paths["merged"] == "/tmp/ws/main/filestore"

    def test_slash_branch(self):
        paths = get_filestore_paths("feature/payments", "/tmp/ws")
        assert paths["merged"] == "/tmp/ws/feature-payments/filestore"


class TestGetTemplateDbName:
    def test_default(self):
        assert get_template_db_name("default") == "odoo_ref_1_default"

    def test_named(self):
        assert get_template_db_name("myproject-v17") == "odoo_ref_1_myproject-v17"

    def test_slash(self):
        assert get_template_db_name("client/prod") == "odoo_ref_1_client-prod"

    def test_default_arg(self):
        assert get_template_db_name() == "odoo_ref_1_default"

    def test_with_instance_id(self):
        assert get_template_db_name("default", instance_id="2") == "odoo_ref_2_default"

    def test_named_with_instance_id(self):
        assert get_template_db_name("myproject-v17", instance_id="3") == "odoo_ref_3_myproject-v17"



import json
import os
import pytest
from oduflow.docker_ops import service_presets
from oduflow.errors import NotFoundError
from oduflow.settings import Settings


@pytest.fixture
def tmp_settings(tmp_path):
    return Settings(
        home=str(tmp_path),
        workspaces_dir=str(tmp_path / "workspaces"),
        port_registry_path=str(tmp_path / "ports.json"),
    )


# ── TestSavePreset ──────────────────────────────────────────────────────


class TestSavePreset:
    def test_save_preset_basic(self, tmp_settings):
        service_presets.save_preset(tmp_settings, "redis", "redis:7", 6379)
        path = service_presets._presets_path(tmp_settings)
        assert os.path.isfile(path)
        with open(path) as fh:
            data = json.load(fh)
        assert "redis" in data
        assert data["redis"]["image"] == "redis:7"
        assert data["redis"]["port"] == 6379
        assert data["redis"]["hostname"] == ""
        assert data["redis"]["env_vars"] == {}

    def test_save_preset_with_env_vars(self, tmp_settings):
        env = {"MEILI_MASTER_KEY": "abc", "MEILI_ENV": "production"}
        service_presets.save_preset(
            tmp_settings, "meili", "getmeili/meilisearch:v1.6", 7700, env_vars=env
        )
        preset = service_presets.get_preset(tmp_settings, "meili")
        assert preset["env_vars"] == env

    def test_save_preset_with_hostname(self, tmp_settings):
        service_presets.save_preset(
            tmp_settings, "redis", "redis:7", 6379, hostname="redis.local"
        )
        preset = service_presets.get_preset(tmp_settings, "redis")
        assert preset["hostname"] == "redis.local"

    def test_save_preset_overwrites(self, tmp_settings):
        service_presets.save_preset(tmp_settings, "redis", "redis:6", 6379)
        service_presets.save_preset(tmp_settings, "redis", "redis:7", 6380)
        preset = service_presets.get_preset(tmp_settings, "redis")
        assert preset["image"] == "redis:7"
        assert preset["port"] == 6380

    def test_save_preset_creates_dir(self, tmp_path):
        nested = tmp_path / "deep" / "nested"
        settings = Settings(
            home=str(nested),
            workspaces_dir=str(nested / "workspaces"),
            port_registry_path=str(nested / "ports.json"),
        )
        assert not nested.exists()
        service_presets.save_preset(settings, "redis", "redis:7", 6379)
        assert os.path.isfile(service_presets._presets_path(settings))

    def test_save_preset_multiple(self, tmp_settings):
        service_presets.save_preset(tmp_settings, "redis", "redis:7", 6379)
        service_presets.save_preset(
            tmp_settings, "meili", "getmeili/meilisearch:v1.6", 7700
        )
        with open(service_presets._presets_path(tmp_settings)) as fh:
            data = json.load(fh)
        assert "redis" in data
        assert "meili" in data


# ── TestListPresets ─────────────────────────────────────────────────────


class TestListPresets:
    def test_list_empty(self, tmp_settings):
        assert service_presets.list_presets(tmp_settings) == []

    def test_list_presets(self, tmp_settings):
        service_presets.save_preset(tmp_settings, "zz-last", "img:1", 1111)
        service_presets.save_preset(tmp_settings, "aa-first", "img:2", 2222)
        result = service_presets.list_presets(tmp_settings)
        assert len(result) == 2
        assert result[0]["name"] == "aa-first"
        assert result[1]["name"] == "zz-last"

    def test_list_preset_format(self, tmp_settings):
        service_presets.save_preset(
            tmp_settings,
            "redis",
            "redis:7",
            6379,
            hostname="r.local",
            env_vars={"K": "V"},
        )
        items = service_presets.list_presets(tmp_settings)
        assert len(items) == 1
        item = items[0]
        for key in ("name", "image", "port", "hostname", "env_vars"):
            assert key in item, f"Missing key: {key}"


# ── TestGetPreset ───────────────────────────────────────────────────────


class TestGetPreset:
    def test_get_preset(self, tmp_settings):
        service_presets.save_preset(
            tmp_settings,
            "redis",
            "redis:7",
            6379,
            hostname="r.local",
            env_vars={"A": "B"},
        )
        preset = service_presets.get_preset(tmp_settings, "redis")
        assert preset["name"] == "redis"
        assert preset["image"] == "redis:7"
        assert preset["port"] == 6379
        assert preset["hostname"] == "r.local"
        assert preset["env_vars"] == {"A": "B"}

    def test_get_preset_not_found(self, tmp_settings):
        with pytest.raises(NotFoundError):
            service_presets.get_preset(tmp_settings, "nonexistent")


# ── TestDeletePreset ────────────────────────────────────────────────────


class TestDeletePreset:
    def test_delete_preset(self, tmp_settings):
        service_presets.save_preset(tmp_settings, "redis", "redis:7", 6379)
        result = service_presets.delete_preset(tmp_settings, "redis")
        assert result == {"name": "redis"}
        with open(service_presets._presets_path(tmp_settings)) as fh:
            data = json.load(fh)
        assert "redis" not in data

    def test_delete_preset_not_found(self, tmp_settings):
        with pytest.raises(NotFoundError):
            service_presets.delete_preset(tmp_settings, "nonexistent")

    def test_delete_preserves_others(self, tmp_settings):
        service_presets.save_preset(tmp_settings, "redis", "redis:7", 6379)
        service_presets.save_preset(tmp_settings, "meili", "meili:1", 7700)
        service_presets.delete_preset(tmp_settings, "redis")
        remaining = service_presets.list_presets(tmp_settings)
        assert len(remaining) == 1
        assert remaining[0]["name"] == "meili"


# ── TestLoadPresetsCorruptFile ──────────────────────────────────────────


class TestLoadPresetsCorruptFile:
    def test_corrupt_json(self, tmp_settings):
        path = service_presets._presets_path(tmp_settings)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as fh:
            fh.write("NOT VALID JSON {{{")
        result = service_presets._load_presets(tmp_settings)
        assert result == {}



import logging
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any

from oduflow.docker_ops.client import get_client
from oduflow.settings import Settings

logger = logging.getLogger("oduflow")


def _calc_cpu_percent(stats: dict) -> float:
    """Calculate CPU usage % from a single stats snapshot."""
    cpu = stats.get("cpu_stats", {})
    precpu = stats.get("precpu_stats", {})

    cpu_delta = cpu.get("cpu_usage", {}).get("total_usage", 0) - precpu.get("cpu_usage", {}).get("total_usage", 0)
    system_delta = cpu.get("system_cpu_usage", 0) - precpu.get("system_cpu_usage", 0)

    if system_delta <= 0 or cpu_delta < 0:
        return 0.0

    num_cpus = cpu.get("online_cpus") or len(cpu.get("cpu_usage", {}).get("percpu_usage", []) or [1])
    return round((cpu_delta / system_delta) * num_cpus * 100.0, 1)


def _get_one_container_stats(container) -> dict[str, Any] | None:
    """Get stats for a single container. Returns None on error."""
    try:
        if container.status != "running":
            return {
                "name": container.name,
                "cpu_percent": 0.0,
                "mem_usage_mb": 0.0,
                "mem_limit_mb": 0.0,
                "mem_percent": 0.0,
            }
        stats = container.stats(stream=False)
        mem = stats.get("memory_stats", {})
        mem_usage = mem.get("usage", 0)
        # Subtract cache if available (actual process memory)
        cache = mem.get("stats", {}).get("cache", 0) if mem.get("stats") else 0
        actual_mem = mem_usage - cache
        mem_limit = mem.get("limit", 0)
        return {
            "name": container.name,
            "cpu_percent": _calc_cpu_percent(stats),
            "mem_usage_mb": round(actual_mem / (1024 * 1024), 1),
            "mem_limit_mb": round(mem_limit / (1024 * 1024), 1),
            "mem_percent": round((actual_mem / mem_limit) * 100.0, 1) if mem_limit > 0 else 0.0,
        }
    except Exception:
        logger.debug("Failed to get stats for %s", container.name, exc_info=True)
        return None


def get_container_stats(settings: Settings) -> list[dict[str, Any]]:
    """Collect CPU/RAM stats for all managed containers in parallel."""
    client = get_client()
    containers = client.containers.list(
        all=True,
        filters={
            "label": [
                f"{settings.managed_label}=true",
                f"{settings.instance_label}={settings.instance_id}"
            ]
        },
    )
    if not containers:
        return []

    results: list[dict[str, Any]] = []
    with ThreadPoolExecutor(max_workers=min(len(containers), 8)) as pool:
        futures = {pool.submit(_get_one_container_stats, c): c for c in containers}
        for future in as_completed(futures):
            stat = future.result()
            if stat is not None:
                results.append(stat)

    results.sort(key=lambda s: s["name"])
    return results


def get_system_stats() -> dict[str, Any]:
    """Read system CPU and memory stats from /proc (Linux only)."""
    result: dict[str, Any] = {
        "mem_total_mb": 0.0,
        "mem_used_mb": 0.0,
        "mem_percent": 0.0,
        "cpu_percent": 0.0,
        "load_avg": [0.0, 0.0, 0.0],
    }

    # Memory from /proc/meminfo
    try:
        info: dict[str, int] = {}
        with open("/proc/meminfo") as f:
            for line in f:
                parts = line.split()
                if len(parts) >= 2:
                    key = parts[0].rstrip(":")
                    info[key] = int(parts[1])  # in kB
        total = info.get("MemTotal", 0)
        available = info.get("MemAvailable", 0)
        used = total - available
        result["mem_total_mb"] = round(total / 1024, 1)
        result["mem_used_mb"] = round(used / 1024, 1)
        result["mem_percent"] = round((used / total) * 100.0, 1) if total > 0 else 0.0
    except Exception:
        logger.debug("Failed to read /proc/meminfo", exc_info=True)

    # Load average
    try:
        load = os.getloadavg()
        result["load_avg"] = [round(v, 2) for v in load]
        result["cpu_count"] = os.cpu_count() or 1
        # Approximate CPU % from 1-minute load average
        result["cpu_percent"] = round((load[0] / result["cpu_count"]) * 100.0, 1)
    except Exception:
        logger.debug("Failed to get load average", exc_info=True)

    return result



import pytest
import docker
from unittest.mock import MagicMock, patch

from oduflow.docker_ops import system_ops, env_ops, odoo_ops
from oduflow.errors import ConflictError, NotFoundError, PrerequisiteNotMetError
from oduflow.settings import Settings

TEST_SETTINGS = Settings(
    external_host="localhost",
    port_range_start=50000,
    port_range_end=50100,
    workspaces_dir="/tmp/flow-test/workspaces",
    home="/tmp/flow-test",
    db_user="odoo",
    db_password="odoo",
    port_registry_path="/tmp/flow-test/ports.json",
)


@pytest.fixture
def mock_docker_client():
    with patch("oduflow.docker_ops.system_ops.get_client") as sys_mock, \
         patch("oduflow.docker_ops.env_ops.get_client") as env_mock, \
         patch("oduflow.docker_ops.odoo_ops.get_client") as odoo_mock:
        client_instance = MagicMock()
        sys_mock.return_value = client_instance
        env_mock.return_value = client_instance
        odoo_mock.return_value = client_instance
        yield client_instance


class TestInitSystem:
    @patch("oduflow.docker_ops.system_ops._copy_file_to_container")
    @patch("oduflow.docker_ops.system_ops.os.path.isfile", return_value=True)
    def test_init_system_fresh(self, mock_isfile, mock_copy, mock_docker_client):
        mock_docker_client.networks.get.side_effect = docker.errors.NotFound("nf")
        mock_docker_client.volumes.get.side_effect = docker.errors.NotFound("nf")

        mock_container = MagicMock()
        mock_docker_client.containers.run.return_value = mock_container

        def get_container(name):
            if name == "oduflow-db":
                c = MagicMock()
                c.status = "running"
                c.exec_run.return_value = (0, b"")
                return c
            raise docker.errors.NotFound("nf")

        mock_docker_client.containers.get.side_effect = get_container

        result = system_ops.init_system(TEST_SETTINGS)

        assert result["status"] == "initialized"
        mock_docker_client.networks.create.assert_called_once()
        mock_docker_client.volumes.create.assert_called_once()

    @patch("oduflow.docker_ops.system_ops._db_exists", return_value=True)
    @patch("oduflow.docker_ops.system_ops._wait_pg_ready")
    def test_init_system_already_initialized(self, mock_pg, mock_db_exists, mock_docker_client):
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.volumes.get.return_value = MagicMock()
        db_container = MagicMock()
        db_container.status = "running"
        mock_docker_client.containers.get.return_value = db_container

        result = system_ops.init_system(TEST_SETTINGS)

        assert result["status"] == "initialized"


class TestDestroySystem:
    def test_destroy_with_active_envs(self, mock_docker_client):
        container = MagicMock()
        container.labels = {"oduflow.branch": "main", "oduflow.managed": "true"}
        container.name = "oduflow-main-odoo"
        mock_docker_client.containers.list.return_value = [container]

        with pytest.raises(ConflictError, match="Active environments exist"):
            system_ops.destroy_system(TEST_SETTINGS)

    def test_destroy_clean(self, mock_docker_client):
        mock_docker_client.containers.list.return_value = []
        db = MagicMock()

        def _get_container(name):
            if name == TEST_SETTINGS.shared_db_container:
                return db
            raise docker.errors.NotFound("not found")

        mock_docker_client.containers.get.side_effect = _get_container
        vol = MagicMock()

        def _get_volume(name):
            if name == TEST_SETTINGS.shared_db_volume:
                return vol
            raise docker.errors.NotFound("not found")

        mock_docker_client.volumes.get.side_effect = _get_volume
        net = MagicMock()
        mock_docker_client.networks.get.return_value = net

        result = system_ops.destroy_system(TEST_SETTINGS)

        assert result["status"] == "destroyed"
        db.stop.assert_called_once()
        db.remove.assert_called_once()
        vol.remove.assert_called_once()
        net.remove.assert_called_once()


class TestCreateEnvironment:
    @patch("oduflow.docker_ops.env_ops._ensure_system_ready")
    @patch("oduflow.docker_ops.env_ops.get_odoo_uid_gid", return_value="100:101")
    @patch("oduflow.docker_ops.env_ops._exec_sql")
    @patch("oduflow.docker_ops.env_ops._db_exists", return_value=True)
    @patch("oduflow.docker_ops.env_ops._mount_filestore")
    @patch("oduflow.docker_ops.env_ops._get_used_ports", return_value=set())
    @patch("oduflow.docker_ops.env_ops.allocate_port", return_value=50000)
    @patch("oduflow.docker_ops.env_ops.subprocess.run")
    @patch("oduflow.docker_ops.env_ops.os.chmod")
    @patch("oduflow.docker_ops.env_ops.os.makedirs")
    @patch("oduflow.docker_ops.env_ops.os.path.exists", return_value=False)
    def test_create(self, mock_exists, mock_makedirs, mock_chmod, mock_run, mock_alloc, mock_used, mock_mount, mock_db_exists, mock_sql, mock_uid_gid, mock_ready, mock_docker_client):
        mock_odoo = MagicMock()
        mock_odoo.exec_run.return_value = (0, b"OK")
        mock_docker_client.containers.run.return_value = mock_odoo
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")

        result = env_ops.create_environment(TEST_SETTINGS, "feature/payments", "https://github.com/org/repo.git", "odoo:15.0")

        assert result["url"] == "http://localhost:50000"
        assert result["database"] == "oduflow_1_feature-payments"
        assert result["odoo_container"] == "oduflow-feature-payments-odoo"
        assert mock_sql.call_count == 2
        mock_docker_client.containers.run.assert_called_once()
        mock_alloc.assert_called_once()

    @patch("oduflow.docker_ops.env_ops._db_exists", return_value=True)
    @patch("oduflow.docker_ops.env_ops._ensure_system_ready")
    def test_create_already_exists(self, mock_ready, mock_db_exists, mock_docker_client):
        existing = MagicMock()
        existing.status = "running"
        existing.ports = {"8069/tcp": [{"HostPort": "50000"}]}
        mock_docker_client.containers.get.return_value = existing

        with pytest.raises(ConflictError, match="already exists"):
            env_ops.create_environment(TEST_SETTINGS, "main", "https://github.com/org/repo.git", "odoo:15.0")

    @patch("oduflow.docker_ops.env_ops._db_exists", return_value=True)
    @patch("oduflow.docker_ops.env_ops._ensure_system_ready")
    def test_create_system_not_ready(self, mock_ready, mock_db_exists, mock_docker_client):
        mock_ready.side_effect = PrerequisiteNotMetError("flow-db not found. Run init_system first.")

        with pytest.raises(PrerequisiteNotMetError, match="init_system"):
            env_ops.create_environment(TEST_SETTINGS, "main", "https://github.com/org/repo.git", "odoo:15.0")


    @patch("oduflow.docker_ops.env_ops._ensure_system_ready")
    @patch("oduflow.docker_ops.env_ops.get_odoo_uid_gid", return_value="100:101")
    @patch("oduflow.docker_ops.env_ops._exec_sql")
    @patch("oduflow.docker_ops.env_ops._db_exists", return_value=False)
    @patch("oduflow.docker_ops.env_ops._mount_filestore")
    @patch("oduflow.docker_ops.env_ops._get_used_ports", return_value=set())
    @patch("oduflow.docker_ops.env_ops.allocate_port", return_value=50001)
    @patch("oduflow.docker_ops.env_ops.subprocess.run")
    @patch("oduflow.docker_ops.env_ops.os.chmod")
    @patch("oduflow.docker_ops.env_ops.os.makedirs")
    @patch("oduflow.docker_ops.env_ops.os.path.exists", return_value=False)
    def test_create_no_template(self, mock_exists, mock_makedirs, mock_chmod, mock_run, mock_alloc, mock_used, mock_mount, mock_db_exists, mock_sql, mock_uid_gid, mock_ready, mock_docker_client):
        mock_odoo = MagicMock()
        mock_odoo.exec_run.return_value = (0, b"OK")
        mock_docker_client.containers.run.return_value = mock_odoo
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")

        result = env_ops.create_environment(TEST_SETTINGS, "feature/no-tpl", "https://github.com/org/repo.git", "odoo:15.0", template_name=None)

        assert result["url"] == "http://localhost:50001"
        assert result["database"] == "oduflow_1_feature-no-tpl"
        # Should create empty DB (no TEMPLATE clause)
        create_db_call = mock_sql.call_args_list[0]
        assert "TEMPLATE" not in create_db_call[0][2]
        # Should NOT mount filestore
        mock_mount.assert_not_called()
        # Should run -i base --stop-after-init
        init_cmd = mock_odoo.exec_run.call_args[0][0]
        assert "-i base" in init_cmd
        assert "--stop-after-init" in init_cmd
        # Should restart after init
        mock_odoo.restart.assert_called_once()


class TestDeleteEnvironment:
    @patch("oduflow.docker_ops.env_ops.release_port")
    @patch("oduflow.docker_ops.env_ops._exec_sql")
    @patch("oduflow.docker_ops.env_ops.shutil.rmtree")
    @patch("oduflow.docker_ops.env_ops.os.path.exists", side_effect=lambda p: ".protected" not in p)
    def test_delete(self, mock_exists, mock_rmtree, mock_sql, mock_release, mock_docker_client):
        container = MagicMock()
        mock_docker_client.containers.get.return_value = container

        env_ops.delete_environment(TEST_SETTINGS, "feature/payments")

        container.stop.assert_called_once()
        container.remove.assert_called_once()
        mock_sql.assert_called_once()
        mock_rmtree.assert_called_once()
        mock_release.assert_called_once_with(TEST_SETTINGS.port_registry_path, "feature/payments")


class TestRestartEnvironment:
    def test_restart(self, mock_docker_client):
        container = MagicMock()
        mock_docker_client.containers.get.return_value = container

        result = env_ops.restart_environment(TEST_SETTINGS, "main")

        assert result["odoo_container"] == "oduflow-main-odoo"
        container.restart.assert_called_once()

    def test_restart_not_found(self, mock_docker_client):
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")

        with pytest.raises(NotFoundError, match="does not exist"):
            env_ops.restart_environment(TEST_SETTINGS, "main")


class TestStopEnvironment:
    def test_stop(self, mock_docker_client):
        container = MagicMock()
        mock_docker_client.containers.get.return_value = container

        result = env_ops.stop_environment(TEST_SETTINGS, "main")

        assert "oduflow-main-odoo" in result["stopped"]
        container.stop.assert_called_once()


class TestStartEnvironment:
    def test_start(self, mock_docker_client):
        db = MagicMock()
        db.status = "running"
        odoo = MagicMock()

        def get_container(name):
            if name == "oduflow-db":
                return db
            return odoo

        mock_docker_client.containers.get.side_effect = get_container

        result = env_ops.start_environment(TEST_SETTINGS, "main")

        assert "oduflow-main-odoo" in result["started"]
        odoo.start.assert_called_once()


class TestGetEnvironmentStatus:
    def test_all_running(self, mock_docker_client):
        odoo = MagicMock()
        odoo.status = "running"
        db = MagicMock()
        db.status = "running"

        def get_container(name):
            if name == "oduflow-db":
                return db
            return odoo

        mock_docker_client.containers.get.side_effect = get_container

        result = env_ops.get_environment_status(TEST_SETTINGS, "main")

        assert result["all_running"] is True
        assert result["db"]["name"] == "oduflow-db"


class TestInstallModules:
    def test_install(self, mock_docker_client):
        container = MagicMock()
        container.exec_run.return_value = (0, b"OK")
        mock_docker_client.containers.get.return_value = container

        result = odoo_ops.install_odoo_modules(TEST_SETTINGS, "main", "sale", "crm")

        assert result["exit_code"] == 0
        args = container.exec_run.call_args[0][0]
        assert "-d oduflow_1_main" in args
        assert "-i sale,crm" in args


class TestUpgradeModules:
    def test_upgrade(self, mock_docker_client):
        container = MagicMock()
        container.exec_run.return_value = (0, b"OK")
        mock_docker_client.containers.get.return_value = container

        result = odoo_ops.upgrade_odoo_modules(TEST_SETTINGS, "main", "sale")

        assert result["exit_code"] == 0
        args = container.exec_run.call_args[0][0]
        assert "-d oduflow_1_main" in args
        assert "-u sale" in args


class TestRunEnvironmentTests:
    def test_run(self, mock_docker_client):
        container = MagicMock()
        container.exec_run.return_value = (0, b"All tests passed")
        mock_docker_client.containers.get.return_value = container

        output = odoo_ops.run_environment_tests(TEST_SETTINGS, "main", "base")

        assert "All tests passed" in output
        args = container.exec_run.call_args[0][0]
        assert "--db_host=oduflow-db" in args
        assert "--database=oduflow_1_main" in args


class TestGetLogs:
    def test_logs(self, mock_docker_client):
        container = MagicMock()
        container.logs.return_value = b"log line 1\nlog line 2"
        mock_docker_client.containers.get.return_value = container

        output = odoo_ops.get_environment_logs(TEST_SETTINGS, "main", 50)

        assert "log line 1" in output
        container.logs.assert_called_with(tail=50, stdout=True, stderr=True)



import pytest
import docker
from unittest.mock import MagicMock, patch

from oduflow.docker_ops import service_ops, system_ops
from oduflow.errors import ConflictError, NotFoundError, PrerequisiteNotMetError
from oduflow.settings import Settings

TEST_SETTINGS = Settings(
    external_host="localhost",
    port_range_start=50000,
    port_range_end=50100,
    workspaces_dir="/tmp/flow-test/workspaces",
    home="/tmp/flow-test",
    db_user="odoo",
    db_password="odoo",
    port_registry_path="/tmp/flow-test/ports.json",
)

TRAEFIK_SETTINGS = Settings(
    routing_mode="traefik",
    base_domain="example.com",
    acme_email="admin@example.com",
    external_host="example.com",
    port_range_start=50000,
    port_range_end=50100,
    workspaces_dir="/tmp/flow-test/workspaces",
    home="/tmp/flow-test",
    db_user="odoo",
    db_password="odoo",
    port_registry_path="/tmp/flow-test/ports.json",
)


@pytest.fixture
def mock_docker_client():
    with patch("oduflow.docker_ops.service_ops.get_client") as svc_mock, \
         patch("oduflow.docker_ops.system_ops.get_client") as sys_mock:
        client_instance = MagicMock()
        svc_mock.return_value = client_instance
        sys_mock.return_value = client_instance
        yield client_instance


class TestCreateService:
    def test_create_port_mode(self, mock_docker_client):
        # Network exists
        mock_docker_client.networks.get.return_value = MagicMock()
        # No existing container
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")
        mock_docker_client.containers.run.return_value = MagicMock()

        result = service_ops.create_service(TEST_SETTINGS, "redis", "redis:7", 6379)

        assert result["name"] == "redis"
        assert result["container_name"] == "oduflow-svc-redis"
        assert result["url"] == "http://localhost:6379"
        assert result["image"] == "redis:7"

        run_kwargs = mock_docker_client.containers.run.call_args
        assert run_kwargs[1]["name"] == "oduflow-svc-redis"
        assert run_kwargs[1]["image"] == "redis:7"
        assert run_kwargs[1]["network"] == "oduflow-net"
        assert run_kwargs[1]["ports"] == {"6379/tcp": 6379}
        assert run_kwargs[1]["labels"]["oduflow.managed"] == "true"
        assert run_kwargs[1]["labels"]["oduflow.service"] == "redis"

    def test_create_traefik_mode(self, mock_docker_client):
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")
        mock_docker_client.containers.run.return_value = MagicMock()

        result = service_ops.create_service(TRAEFIK_SETTINGS, "meilisearch", "getmeili/meilisearch:v1.6", 7700)

        assert result["url"] == "https://meilisearch.example.com"

        run_kwargs = mock_docker_client.containers.run.call_args
        labels = run_kwargs[1]["labels"]
        assert labels["traefik.enable"] == "true"
        assert labels["traefik.http.routers.oduflow-svc-meilisearch.rule"] == "Host(`meilisearch.example.com`)"
        assert labels["traefik.http.services.oduflow-svc-meilisearch.loadbalancer.server.port"] == "7700"
        # No port mapping in traefik mode
        assert "ports" not in run_kwargs[1]

    def test_create_traefik_custom_hostname(self, mock_docker_client):
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")
        mock_docker_client.containers.run.return_value = MagicMock()

        result = service_ops.create_service(
            TRAEFIK_SETTINGS, "redis", "redis:7", 6379, hostname="my-redis.example.com"
        )

        assert result["url"] == "https://my-redis.example.com"
        labels = mock_docker_client.containers.run.call_args[1]["labels"]
        assert labels["traefik.http.routers.oduflow-svc-redis.rule"] == "Host(`my-redis.example.com`)"

    def test_create_with_env_vars(self, mock_docker_client):
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")
        mock_docker_client.containers.run.return_value = MagicMock()

        env = {"MEILI_MASTER_KEY": "abc123", "MEILI_ENV": "production"}
        result = service_ops.create_service(TEST_SETTINGS, "meili", "getmeili/meilisearch:v1.6", 7700, env_vars=env)

        run_kwargs = mock_docker_client.containers.run.call_args
        assert run_kwargs[1]["environment"] == env

    def test_create_without_env_vars(self, mock_docker_client):
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")
        mock_docker_client.containers.run.return_value = MagicMock()

        service_ops.create_service(TEST_SETTINGS, "redis", "redis:7", 6379)

        run_kwargs = mock_docker_client.containers.run.call_args
        assert "environment" not in run_kwargs[1]

    def test_create_conflict(self, mock_docker_client):
        mock_docker_client.networks.get.return_value = MagicMock()
        existing = MagicMock()
        existing.status = "running"
        mock_docker_client.containers.get.return_value = existing

        with pytest.raises(ConflictError, match="already exists"):
            service_ops.create_service(TEST_SETTINGS, "redis", "redis:7", 6379)

        mock_docker_client.containers.run.assert_not_called()

    def test_create_network_missing(self, mock_docker_client):
        mock_docker_client.networks.get.side_effect = docker.errors.NotFound("nf")

        with pytest.raises(PrerequisiteNotMetError, match="init_system"):
            service_ops.create_service(TEST_SETTINGS, "redis", "redis:7", 6379)

        mock_docker_client.containers.run.assert_not_called()


class TestDeleteService:
    def test_delete(self, mock_docker_client):
        container = MagicMock()
        mock_docker_client.containers.get.return_value = container

        result = service_ops.delete_service(TEST_SETTINGS, "redis")

        assert result["name"] == "redis"
        assert result["container_name"] == "oduflow-svc-redis"
        container.stop.assert_called_once()
        container.remove.assert_called_once_with(v=True)

    def test_delete_not_found(self, mock_docker_client):
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")

        with pytest.raises(NotFoundError, match="Service 'redis' not found"):
            service_ops.delete_service(TEST_SETTINGS, "redis")


class TestListServices:
    def test_list_with_services_port_mode(self, mock_docker_client):
        container = MagicMock()
        container.labels = {
            "oduflow.managed": "true",
            "oduflow.service": "redis",
        }
        container.name = "oduflow-svc-redis"
        container.status = "running"
        container.image.tags = ["redis:7"]
        container.attrs = {
            "NetworkSettings": {
                "Ports": {"6379/tcp": [{"HostIp": "0.0.0.0", "HostPort": "6379"}]}
            },
            "Config": {"Env": ["REDIS_PASSWORD=secret", "PATH=/usr/bin", "HOME=/root"]},
        }
        mock_docker_client.containers.list.return_value = [container]

        result = service_ops.list_services(TEST_SETTINGS)

        assert len(result) == 1
        svc = result[0]
        assert svc["name"] == "redis"
        assert svc["container_name"] == "oduflow-svc-redis"
        assert svc["image"] == "redis:7"
        assert svc["status"] == "running"
        assert svc["port"] == 6379
        assert svc["url"] == "http://localhost:6379"
        # Env vars: system vars filtered out, only custom remain
        assert svc["env_vars"] == {"REDIS_PASSWORD": "secret"}

    def test_list_with_services_traefik_mode(self, mock_docker_client):
        container = MagicMock()
        container.labels = {
            "oduflow.managed": "true",
            "oduflow.service": "meili",
            "traefik.enable": "true",
            "traefik.http.routers.oduflow-svc-meili.rule": "Host(`meili.example.com`)",
            "traefik.http.services.oduflow-svc-meili.loadbalancer.server.port": "7700",
        }
        container.name = "oduflow-svc-meili"
        container.status = "running"
        container.image.tags = ["getmeili/meilisearch:v1.6"]
        container.attrs = {"Config": {"Env": []}}
        mock_docker_client.containers.list.return_value = [container]

        result = service_ops.list_services(TRAEFIK_SETTINGS)

        assert len(result) == 1
        svc = result[0]
        assert svc["url"] == "https://meili.example.com"
        assert svc["port"] == 7700

    def test_list_empty(self, mock_docker_client):
        mock_docker_client.containers.list.return_value = []

        result = service_ops.list_services(TEST_SETTINGS)

        assert result == []

    def test_list_skips_non_service_containers(self, mock_docker_client):
        # Container with managed label but no service label (e.g. an odoo env)
        container = MagicMock()
        container.labels = {"oduflow.managed": "true", "oduflow.branch": "main"}
        mock_docker_client.containers.list.return_value = [container]

        result = service_ops.list_services(TEST_SETTINGS)

        assert result == []

    def test_list_env_var_filtering(self, mock_docker_client):
        container = MagicMock()
        container.labels = {"oduflow.managed": "true", "oduflow.service": "meili"}
        container.name = "oduflow-svc-meili"
        container.status = "running"
        container.image.tags = ["getmeili/meilisearch:v1.6"]
        container.attrs = {
            "NetworkSettings": {"Ports": {}},
            "Config": {
                "Env": [
                    "MEILI_MASTER_KEY=abc",
                    "MEILI_ENV=production",
                    "PATH=/usr/local/bin:/usr/bin",
                    "HOME=/root",
                    "HOSTNAME=abc123",
                    "TERM=xterm",
                    "LANG=en_US.UTF-8",
                    "LC_ALL=en_US.UTF-8",
                ]
            },
        }
        mock_docker_client.containers.list.return_value = [container]

        result = service_ops.list_services(TEST_SETTINGS)

        env = result[0]["env_vars"]
        assert env == {"MEILI_MASTER_KEY": "abc", "MEILI_ENV": "production"}
        # System vars must be excluded
        for sys_key in ("PATH", "HOME", "HOSTNAME", "TERM", "LANG", "LC_ALL"):
            assert sys_key not in env

    def test_list_image_fallback(self, mock_docker_client):
        """When image.tags is empty, fall back to Config.Image."""
        container = MagicMock()
        container.labels = {"oduflow.managed": "true", "oduflow.service": "redis"}
        container.name = "oduflow-svc-redis"
        container.status = "running"
        container.image.tags = []
        container.attrs = {
            "NetworkSettings": {"Ports": {}},
            "Config": {"Image": "redis:7-alpine", "Env": []},
        }
        mock_docker_client.containers.list.return_value = [container]

        result = service_ops.list_services(TEST_SETTINGS)

        assert result[0]["image"] == "unknown"

    def test_list_port_no_mappings(self, mock_docker_client):
        """Port key exists but no host mappings."""
        container = MagicMock()
        container.labels = {"oduflow.managed": "true", "oduflow.service": "redis"}
        container.name = "oduflow-svc-redis"
        container.status = "exited"
        container.image.tags = ["redis:7"]
        container.attrs = {
            "NetworkSettings": {"Ports": {"6379/tcp": None}},
            "Config": {"Env": []},
        }
        mock_docker_client.containers.list.return_value = [container]

        result = service_ops.list_services(TEST_SETTINGS)

        svc = result[0]
        assert svc["port"] == 6379
        assert svc["url"] is None


class TestUpdateService:
    def test_update_port_mode(self, mock_docker_client):
        """Update pulls the image, removes old container, and re-creates with same settings."""
        container = MagicMock()
        container.image.tags = ["redis:7"]
        container.labels = {"oduflow.managed": "true", "oduflow.service": "redis"}
        container.attrs = {
            "NetworkSettings": {"Ports": {"6379/tcp": [{"HostIp": "0.0.0.0", "HostPort": "6379"}]}},
            "Config": {"Env": ["REDIS_PASSWORD=secret", "PATH=/usr/bin", "HOME=/root"]},
        }

        # First get() returns the existing container, second get() (inside create_service)
        # raises NotFound (container was removed)
        mock_docker_client.containers.get.side_effect = [
            container,  # update_service lookup
            docker.errors.NotFound("nf"),  # create_service conflict check
        ]
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.run.return_value = MagicMock()

        result = service_ops.update_service(TEST_SETTINGS, "redis")

        assert result["name"] == "redis"
        assert result["image"] == "redis:7"
        assert result["url"] == "http://localhost:6379"

        # Verify pull was called
        mock_docker_client.images.pull.assert_called_once_with("redis:7")

        # Verify old container was stopped and removed
        container.stop.assert_called_once()
        container.remove.assert_called_once_with(v=True)

        # Verify re-creation with same env vars (only non-system)
        run_kwargs = mock_docker_client.containers.run.call_args
        assert run_kwargs[1]["environment"] == {"REDIS_PASSWORD": "secret"}

    def test_update_traefik_mode(self, mock_docker_client):
        container = MagicMock()
        container.image.tags = ["getmeili/meilisearch:v1.6"]
        container.labels = {
            "oduflow.managed": "true",
            "oduflow.service": "meili",
            "traefik.http.routers.oduflow-svc-meili.rule": "Host(`meili.example.com`)",
            "traefik.http.services.oduflow-svc-meili.loadbalancer.server.port": "7700",
        }
        container.attrs = {"Config": {"Env": ["MEILI_MASTER_KEY=abc"]}}

        mock_docker_client.containers.get.side_effect = [
            container,
            docker.errors.NotFound("nf"),
        ]
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.run.return_value = MagicMock()

        result = service_ops.update_service(TRAEFIK_SETTINGS, "meili")

        assert result["url"] == "https://meili.example.com"
        mock_docker_client.images.pull.assert_called_once_with("getmeili/meilisearch:v1.6")

        # Verify traefik labels are restored
        run_kwargs = mock_docker_client.containers.run.call_args
        labels = run_kwargs[1]["labels"]
        assert labels["traefik.http.routers.oduflow-svc-meili.rule"] == "Host(`meili.example.com`)"

    def test_update_no_env_vars(self, mock_docker_client):
        """When the container has no custom env vars, env_vars=None is passed to create."""
        container = MagicMock()
        container.image.tags = ["redis:7"]
        container.labels = {"oduflow.managed": "true", "oduflow.service": "redis"}
        container.attrs = {
            "NetworkSettings": {"Ports": {"6379/tcp": [{"HostIp": "0.0.0.0", "HostPort": "6379"}]}},
            "Config": {"Env": ["PATH=/usr/bin", "HOME=/root"]},
        }

        mock_docker_client.containers.get.side_effect = [
            container,
            docker.errors.NotFound("nf"),
        ]
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.run.return_value = MagicMock()

        service_ops.update_service(TEST_SETTINGS, "redis")

        run_kwargs = mock_docker_client.containers.run.call_args
        assert "environment" not in run_kwargs[1]

    def test_update_not_found(self, mock_docker_client):
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")

        with pytest.raises(NotFoundError, match="Service 'redis' not found"):
            service_ops.update_service(TEST_SETTINGS, "redis")

    def test_update_image_fallback_to_config(self, mock_docker_client):
        """When image.tags is empty, fall back to Config.Image."""
        container = MagicMock()
        container.image.tags = []
        container.labels = {"oduflow.managed": "true", "oduflow.service": "redis"}
        container.attrs = {
            "NetworkSettings": {"Ports": {"6379/tcp": [{"HostIp": "0.0.0.0", "HostPort": "6379"}]}},
            "Config": {"Image": "redis:7-alpine", "Env": []},
        }

        mock_docker_client.containers.get.side_effect = [
            container,
            docker.errors.NotFound("nf"),
        ]
        mock_docker_client.networks.get.return_value = MagicMock()
        mock_docker_client.containers.run.return_value = MagicMock()

        result = service_ops.update_service(TEST_SETTINGS, "redis")

        assert result["image"] == "redis:7-alpine"
        mock_docker_client.images.pull.assert_called_once_with("redis:7-alpine")


class TestGetServiceLogs:
    def test_logs(self, mock_docker_client):
        container = MagicMock()
        container.logs.return_value = b"2025-01-01T00:00:00Z log line 1\n2025-01-01T00:00:01Z log line 2"
        mock_docker_client.containers.get.return_value = container

        output = service_ops.get_service_logs(TEST_SETTINGS, "redis", 50)

        assert "log line 1" in output
        assert "log line 2" in output
        container.logs.assert_called_with(tail=50, timestamps=True)

    def test_logs_default_lines(self, mock_docker_client):
        container = MagicMock()
        container.logs.return_value = b"line"
        mock_docker_client.containers.get.return_value = container

        service_ops.get_service_logs(TEST_SETTINGS, "redis")

        container.logs.assert_called_with(tail=100, timestamps=True)

    def test_logs_not_found(self, mock_docker_client):
        mock_docker_client.containers.get.side_effect = docker.errors.NotFound("nf")

        with pytest.raises(NotFoundError, match="Service 'redis' not found"):
            service_ops.get_service_logs(TEST_SETTINGS, "redis")


class TestDestroyBlockedByServices:
    def test_destroy_with_active_services(self, mock_docker_client):
        container = MagicMock()
        container.labels = {"oduflow.service": "redis", "oduflow.managed": "true"}
        container.name = "oduflow-svc-redis"
        mock_docker_client.containers.list.return_value = [container]

        with pytest.raises(ConflictError, match="Active environments/services exist"):
            system_ops.destroy_system(TEST_SETTINGS)



import logging
from typing import Any

import docker

from oduflow.docker_ops.client import get_client
from oduflow.errors import NotFoundError
from oduflow.naming import get_db_name, get_resource_name
from oduflow.settings import Settings

logger = logging.getLogger("oduflow")


def run_environment_tests(settings: Settings, branch_name: str, modules: str) -> str:
    client = get_client()
    odoo_container_name = get_resource_name(branch_name, "odoo", settings.prefix)
    env_db = get_db_name(branch_name, settings.instance_id)

    try:
        container = client.containers.get(odoo_container_name)
    except docker.errors.NotFound:
        raise NotFoundError(
            f"Environment '{branch_name}' does not exist. Use create_environment first."
        )

    cmd = (
        f"odoo --test-enable --stop-after-init -i {modules} "
        f"--db_host={settings.shared_db_container} "
        f"-r {settings.db_user} -w {settings.db_password} "
        f"--database={env_db}"
    )

    logger.info(
        "Running tests",
        extra={"branch": branch_name, "modules": modules},
    )
    exit_code, output = container.exec_run(cmd)

    if isinstance(output, bytes):
        return output.decode("utf-8")
    return str(output)


def get_environment_logs(settings: Settings, branch_name: str, n_lines: int = 100) -> str:
    client = get_client()
    odoo_container_name = get_resource_name(branch_name, "odoo", settings.prefix)

    try:
        container = client.containers.get(odoo_container_name)
        logs = container.logs(tail=n_lines, stdout=True, stderr=True)
        if isinstance(logs, bytes):
            return logs.decode("utf-8")
        return str(logs)
    except docker.errors.NotFound:
        raise NotFoundError(
            f"Environment '{branch_name}' does not exist. Use create_environment first."
        )


def _run_odoo_module_command(
    settings: Settings, branch_name: str, flag: str, *modules: str
) -> dict[str, Any]:
    if not modules:
        raise ValueError("At least one module name is required.")

    client = get_client()
    odoo_container_name = get_resource_name(branch_name, "odoo", settings.prefix)
    env_db = get_db_name(branch_name, settings.instance_id)

    try:
        container = client.containers.get(odoo_container_name)
    except docker.errors.NotFound:
        raise NotFoundError(
            f"Environment '{branch_name}' does not exist. Use create_environment first."
        )

    modules_str = ",".join(modules)
    cmd = f"/entrypoint.sh odoo -d {env_db} --stop-after-init --no-http {flag} {modules_str}"

    action = "Installing" if flag == "-i" else "Upgrading"
    logger.info(
        "%s modules", action,
        extra={"branch": branch_name, "modules": modules_str},
    )
    exit_code, output = container.exec_run(cmd)
    output_str = output.decode("utf-8") if isinstance(output, bytes) else str(output)

    return {
        "modules": list(modules),
        "exit_code": exit_code,
        "output": output_str,
    }


def upgrade_odoo_modules(settings: Settings, branch_name: str, *modules: str) -> dict[str, Any]:
    return _run_odoo_module_command(settings, branch_name, "-u", *modules)


def install_odoo_modules(settings: Settings, branch_name: str, *modules: str) -> dict[str, Any]:
    return _run_odoo_module_command(settings, branch_name, "-i", *modules)


def exec_in_environment(
    settings: Settings, branch_name: str, command: str, user: str = "odoo"
) -> dict[str, Any]:
    client = get_client()
    odoo_container_name = get_resource_name(branch_name, "odoo", settings.prefix)

    try:
        container = client.containers.get(odoo_container_name)
    except docker.errors.NotFound:
        raise NotFoundError(
            f"Environment '{branch_name}' does not exist. Use create_environment first."
        )

    logger.info(
        "Executing command in environment",
        extra={"branch": branch_name, "command": command, "user": user},
    )
    exit_code, output = container.exec_run(command, user=user)
    output_str = output.decode("utf-8") if isinstance(output, bytes) else str(output)

    return {
        "exit_code": exit_code,
        "output": output_str,
    }



import io
import json
import logging
import os
import pathlib
import shutil
import subprocess
import tarfile
import time

import docker
from docker import DockerClient

from oduflow.docker_ops.client import get_client, get_odoo_uid_gid
from oduflow.errors import ConflictError, ExternalCommandError, NotFoundError, PrerequisiteNotMetError
from oduflow.naming import get_db_name, get_template_db_name
from oduflow.settings import Settings

logger = logging.getLogger("oduflow")

_PACKAGE_ROOT = pathlib.Path(__file__).resolve().parents[1]
_BUNDLED_PG_CONF = _PACKAGE_ROOT / "templates" / "postgresql.conf"
_BUNDLED_ODOO_CONF = _PACKAGE_ROOT / "templates" / "odoo.conf"
_ETC_DIR = pathlib.Path("/etc/oduflow")


def _normalize_extra_addons(raw_addons, fallback_branch: str) -> dict[str, str]:
    """Convert old list format or new dict format to {name: branch} dict."""
    if isinstance(raw_addons, dict):
        return raw_addons
    if isinstance(raw_addons, list):
        return {name: fallback_branch for name in raw_addons}
    return {}


def _resolve_conf(name: str) -> pathlib.Path:
    """Return /etc/oduflow/{name} if present, otherwise the bundled copy."""
    etc_path = _ETC_DIR / name
    if etc_path.is_file():
        return etc_path
    bundled = _PACKAGE_ROOT / "templates" / name
    return bundled


def _ensure_traefik(client: DockerClient, settings: Settings) -> None:
    if settings.routing_mode != "traefik":
        return

    system_labels = {settings.managed_label: "true", settings.system_label: "true"}

    try:
        client.volumes.get(settings.traefik_acme_volume)
    except docker.errors.NotFound:
        client.volumes.create(settings.traefik_acme_volume, labels=system_labels)
        logger.info("Created volume %s", settings.traefik_acme_volume)

    try:
        t = client.containers.get(settings.traefik_container)
        if t.status != "running":
            t.start()
        return
    except docker.errors.NotFound:
        pass

    dynamic_dir = "/etc/oduflow/traefik"
    os.makedirs(dynamic_dir, exist_ok=True)

    client.containers.run(
        "traefik:v3.6",
        name=settings.traefik_container,
        detach=True,
        network=settings.shared_network,
        ports={"80/tcp": 80, "443/tcp": 443},
        extra_hosts={"host.docker.internal": "host-gateway"},
        volumes={
            "/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "ro"},
            settings.traefik_acme_volume: {"bind": "/acme", "mode": "rw"},
            dynamic_dir: {"bind": "/etc/traefik/dynamic/", "mode": "ro"},
        },
        command=[
            "--log.level=INFO",
            "--providers.docker=true",
            "--providers.docker.exposedbydefault=false",
            f"--providers.docker.network={settings.shared_network}",
            "--providers.file.directory=/etc/traefik/dynamic/",
            "--providers.file.watch=true",
            "--entrypoints.web.address=:80",
            "--entrypoints.websecure.address=:443",
            "--entrypoints.web.http.redirections.entrypoint.to=websecure",
            "--entrypoints.web.http.redirections.entrypoint.scheme=https",
            "--certificatesresolvers.le.acme.httpchallenge=true",
            "--certificatesresolvers.le.acme.httpchallenge.entrypoint=web",
            f"--certificatesresolvers.le.acme.email={settings.acme_email}",
            "--certificatesresolvers.le.acme.storage=/acme/acme.json",
        ],
        labels=system_labels,
        restart_policy={"Name": "unless-stopped"},
    )
    logger.info("Created container %s", settings.traefik_container)


def _destroy_traefik(client: DockerClient, settings: Settings, removed: list[str]) -> None:
    try:
        t = client.containers.get(settings.traefik_container)
        t.stop()
        t.remove(v=True)
        removed.append(settings.traefik_container)
    except docker.errors.NotFound:
        pass

    try:
        v = client.volumes.get(settings.traefik_acme_volume)
        v.remove()
        removed.append(settings.traefik_acme_volume)
    except docker.errors.NotFound:
        pass


def _wait_pg_ready(client: DockerClient, settings: Settings, timeout: int = 30) -> None:
    container = client.containers.get(settings.shared_db_container)
    for i in range(timeout):
        exit_code, _ = container.exec_run(["pg_isready", "-U", settings.db_user])
        if exit_code == 0:
            exit_code2, _ = container.exec_run(
                ["psql", "-U", settings.db_user, "-d", "postgres", "-tAc", "SELECT 1;"]
            )
            if exit_code2 == 0:
                return
        time.sleep(1)
    raise PrerequisiteNotMetError(
        f"PostgreSQL did not become ready within {timeout}s"
    )


def _exec_sql(client: DockerClient, settings: Settings, sql: str, db: str = "postgres") -> str:
    container = client.containers.get(settings.shared_db_container)
    exit_code, output = container.exec_run(
        ["psql", "-U", settings.db_user, "-d", db, "-tAc", sql]
    )
    result = output.decode("utf-8") if isinstance(output, bytes) else str(output)
    if exit_code != 0:
        raise ExternalCommandError("psql", exit_code, result)
    return result.strip()


def _db_exists(client: DockerClient, settings: Settings, db_name: str) -> bool:
    result = _exec_sql(
        client,
        settings,
        f"SELECT 1 FROM pg_database WHERE datname='{db_name}';",
    )
    return result == "1"


def _is_text_dump(path: str) -> bool:
    try:
        with open(path, "rb") as f:
            header = f.read(5)
        return header != b"PGDMP"
    except OSError:
        return False


def _copy_file_to_container(container: docker.models.containers.Container, src_path: str, dest_dir: str) -> None:
    with open(src_path, "rb") as f:
        data = f.read()
    tar_stream = io.BytesIO()
    with tarfile.open(fileobj=tar_stream, mode="w") as tar:
        info = tarfile.TarInfo(name=os.path.basename(src_path))
        info.size = len(data)
        tar.addfile(info, io.BytesIO(data))
    tar_stream.seek(0)
    container.put_archive(dest_dir, tar_stream)


def _ensure_iptables_accept(client: DockerClient, network_name: str) -> None:
    try:
        net = client.networks.get(network_name)
        bridge_iface = "br-" + net.id[:12]
    except docker.errors.NotFound:
        return
    try:
        subprocess.run(
            ["iptables", "-C", "INPUT", "-i", bridge_iface, "-j", "ACCEPT"],
            check=True, capture_output=True,
        )
        logger.debug("iptables ACCEPT rule already exists for %s", bridge_iface)
    except (subprocess.CalledProcessError, FileNotFoundError):
        try:
            subprocess.run(
                ["iptables", "-I", "INPUT", "-i", bridge_iface, "-j", "ACCEPT"],
                check=True, capture_output=True,
            )
            logger.info("Added iptables ACCEPT rule for interface %s", bridge_iface)
        except (subprocess.CalledProcessError, FileNotFoundError) as exc:
            logger.warning("Could not add iptables rule for %s: %s", bridge_iface, exc)


def init_system(
    settings: Settings,
) -> dict[str, str]:
    client = get_client()
    logger.info("Initializing system")

    system_labels = {settings.managed_label: "true", settings.system_label: "true"}

    try:
        client.networks.get(settings.shared_network)
    except docker.errors.NotFound:
        client.networks.create(settings.shared_network, labels=system_labels)
        logger.info("Created network %s", settings.shared_network)

    _ensure_iptables_accept(client, settings.shared_network)

    _ensure_traefik(client, settings)

    try:
        client.volumes.get(settings.shared_db_volume)
    except docker.errors.NotFound:
        client.volumes.create(settings.shared_db_volume, labels=system_labels)
        logger.info("Created volume %s", settings.shared_db_volume)

    try:
        db_container = client.containers.get(settings.shared_db_container)
        if db_container.status != "running":
            db_container.start()
    except docker.errors.NotFound:
        client.containers.run(
            settings.postgres_image,
            name=settings.shared_db_container,
            detach=True,
            network=settings.shared_network,
            volumes={
                settings.shared_db_volume: {"bind": "/var/lib/postgresql/data", "mode": "rw"},
                str(_resolve_conf("postgresql.conf")): {"bind": "/etc/postgresql/postgresql.conf", "mode": "ro"},
            },
            command=["postgres", "-c", "config_file=/etc/postgresql/postgresql.conf"],
            environment={
                "POSTGRES_USER": settings.db_user,
                "POSTGRES_PASSWORD": settings.db_password,
            },
            labels=system_labels,
            restart_policy={"Name": "unless-stopped"},
        )
        logger.info("Created container %s", settings.shared_db_container)

    _wait_pg_ready(client, settings)

    logger.info("System initialized")
    return {"status": "initialized"}


def reload_template(
    settings: Settings,
    template_name: str,
    dump_path: str | None = None,
) -> dict[str, str]:
    client = get_client()
    tpl_db = get_template_db_name(template_name, settings.instance_id)
    resolved_dump = dump_path or settings.get_template_sql_path(template_name)

    if not os.path.isfile(resolved_dump):
        raise NotFoundError(f"Dump file not found: {resolved_dump}")

    _wait_pg_ready(client, settings)

    if _db_exists(client, settings, tpl_db):
        _exec_sql(
            client,
            settings,
            f"UPDATE pg_database SET datistemplate=false WHERE datname='{tpl_db}';",
        )
        _exec_sql(client, settings, f'DROP DATABASE "{tpl_db}" WITH (FORCE);')
        logger.info("Dropped template DB %s", tpl_db)

    _exec_sql(client, settings, f'CREATE DATABASE "{tpl_db}";')

    db_container = client.containers.get(settings.shared_db_container)
    tmp_name = os.path.basename(resolved_dump)

    _copy_file_to_container(db_container, resolved_dump, "/tmp")

    use_psql = _is_text_dump(resolved_dump)

    if use_psql:
        restore_cmd = ["psql", "-U", settings.db_user, "-d", tpl_db, "-f", f"/tmp/{tmp_name}"]
    else:
        restore_cmd = ["pg_restore", "-U", settings.db_user, "-d", tpl_db, f"/tmp/{tmp_name}"]

    logger.info("DB restore started, template_db=%s, dump=%s", tpl_db, resolved_dump)
    restore_start = time.monotonic()

    exit_code, output = db_container.exec_run(restore_cmd)

    restore_elapsed = time.monotonic() - restore_start

    if exit_code != 0:
        logger.error("DB restore failed after %.1fs", restore_elapsed)
        msg = output.decode("utf-8") if isinstance(output, bytes) else str(output)
        cmd_name = "psql" if use_psql else "pg_restore"
        raise ExternalCommandError(cmd_name, exit_code, msg)

    logger.info("DB restore finished in %.1fs", restore_elapsed)

    _exec_sql(
        client,
        settings,
        f"UPDATE pg_database SET datistemplate=true WHERE datname='{tpl_db}';",
    )

    logger.info("Template DB reloaded, template_db=%s, restore_time=%.1fs", tpl_db, restore_elapsed)
    return {"status": "reloaded", "template_db": tpl_db, "restore_seconds": round(restore_elapsed, 1)}


def init_template(
    settings: Settings,
    template_name: str,
    odoo_image: str = "odoo:17.0",
    modules: str = "base",
    force: bool = False,
) -> dict[str, str]:
    template_sql_path = settings.get_template_sql_path(template_name)
    template_filestore_path = settings.get_template_filestore_path(template_name)

    existing_dump = os.path.exists(template_sql_path)
    existing_filestore = (
        os.path.exists(template_filestore_path)
        and any(True for _ in pathlib.Path(template_filestore_path).rglob("*") if _.is_file())
    )

    if (existing_dump or existing_filestore) and not force:
        parts = []
        if existing_dump:
            parts.append(f"dump.sql ({template_sql_path})")
        if existing_filestore:
            parts.append(f"filestore ({template_filestore_path})")
        raise RuntimeError(
            f"Existing data found: {', '.join(parts)}. "
            "Use --force to overwrite."
        )

    if force:
        if existing_dump:
            os.remove(template_sql_path)
            logger.info("Removed existing %s", template_sql_path)
        if existing_filestore:
            shutil.rmtree(template_filestore_path)
            logger.info("Removed existing %s", template_filestore_path)

    client = get_client()
    logger.info("Generating template dump from clean Odoo", extra={"image": odoo_image, "modules": modules})

    system_labels = {settings.managed_label: "true", settings.system_label: "true"}

    try:
        client.networks.get(settings.shared_network)
    except docker.errors.NotFound:
        client.networks.create(settings.shared_network, labels=system_labels)
        logger.info("Created network %s", settings.shared_network)

    try:
        client.volumes.get(settings.shared_db_volume)
    except docker.errors.NotFound:
        client.volumes.create(settings.shared_db_volume, labels=system_labels)
        logger.info("Created volume %s", settings.shared_db_volume)

    try:
        db_container = client.containers.get(settings.shared_db_container)
        if db_container.status != "running":
            db_container.start()
    except docker.errors.NotFound:
        client.containers.run(
            settings.postgres_image,
            name=settings.shared_db_container,
            detach=True,
            network=settings.shared_network,
            volumes={
                settings.shared_db_volume: {"bind": "/var/lib/postgresql/data", "mode": "rw"},
                str(_resolve_conf("postgresql.conf")): {"bind": "/etc/postgresql/postgresql.conf", "mode": "ro"},
            },
            command=["postgres", "-c", "config_file=/etc/postgresql/postgresql.conf"],
            environment={
                "POSTGRES_USER": settings.db_user,
                "POSTGRES_PASSWORD": settings.db_password,
            },
            labels=system_labels,
            restart_policy={"Name": "unless-stopped"},
        )
        logger.info("Created container %s", settings.shared_db_container)

    _wait_pg_ready(client, settings)

    build_db = "odoo_ref_build"
    temp_container_name = "flow-template-builder"

    if _db_exists(client, settings, build_db):
        _exec_sql(client, settings, f"DROP DATABASE {build_db} WITH (FORCE);")

    _exec_sql(client, settings, f"CREATE DATABASE {build_db};")
    logger.info("Created temporary database %s", build_db)

    try:
        old = client.containers.get(temp_container_name)
        old.remove(force=True)
    except docker.errors.NotFound:
        pass

    logger.info("Starting Odoo container for base init (image=%s, modules=%s)", odoo_image, modules)
    init_start = time.monotonic()

    volumes = {}
    odoo_conf = _resolve_conf("odoo.conf")
    if odoo_conf.exists():
        volumes[str(odoo_conf)] = {"bind": "/etc/odoo/odoo.conf", "mode": "ro"}

    temp_container = client.containers.run(
        odoo_image,
        name=temp_container_name,
        detach=True,
        network=settings.shared_network,
        environment={
            "HOST": settings.shared_db_container,
            "USER": settings.db_user,
            "PASSWORD": settings.db_password,
        },
        volumes=volumes,
        command=f"odoo -d {build_db} -i {modules} --stop-after-init --without-demo=all",
        labels={settings.managed_label: "true"},
    )

    exit_info = temp_container.wait(timeout=600)
    init_elapsed = time.monotonic() - init_start
    exit_code = exit_info.get("StatusCode", -1)

    if exit_code != 0:
        logs = temp_container.logs(tail=50).decode("utf-8", errors="replace")
        temp_container.remove(v=True)
        _exec_sql(client, settings, f"DROP DATABASE IF EXISTS {build_db} WITH (FORCE);")
        raise ExternalCommandError(
            "odoo --stop-after-init", exit_code,
            f"Odoo init failed after {init_elapsed:.1f}s.\nLast logs:\n{logs}",
        )

    logger.info("Odoo init completed in %.1fs", init_elapsed)

    os.makedirs(os.path.dirname(template_sql_path), exist_ok=True)

    db_container = client.containers.get(settings.shared_db_container)
    dump_cmd = [
        "pg_dump", "-U", settings.db_user, "-Fp", "-f", f"/tmp/{build_db}.dump", build_db,
    ]
    exit_code_dump, output_dump = db_container.exec_run(dump_cmd)
    if exit_code_dump != 0:
        msg = output_dump.decode("utf-8") if isinstance(output_dump, bytes) else str(output_dump)
        temp_container.remove(v=True)
        _exec_sql(client, settings, f"DROP DATABASE IF EXISTS {build_db} WITH (FORCE);")
        raise ExternalCommandError("pg_dump", exit_code_dump, msg)

    logger.info("pg_dump completed, extracting dump file")

    chunks, _ = db_container.get_archive(f"/tmp/{build_db}.dump")
    raw = b"".join(chunks)
    tar_stream = io.BytesIO(raw)
    with tarfile.open(fileobj=tar_stream, mode="r") as tar:
        member = tar.getmembers()[0]
        f = tar.extractfile(member)
        if f is None:
            raise ExternalCommandError("get_archive", -1, "Could not extract dump from tar")
        with open(template_sql_path, "wb") as out:
            out.write(f.read())

    logger.info("Dump saved to %s", template_sql_path)

    if os.path.exists(template_filestore_path):
        shutil.rmtree(template_filestore_path)
    os.makedirs(template_filestore_path, exist_ok=True)

    odoo_data_container_path = "/var/lib/odoo/.local/share/Odoo"
    try:
        chunks_fs, _ = temp_container.get_archive(odoo_data_container_path)
        raw_fs = b"".join(chunks_fs)
        tar_fs = io.BytesIO(raw_fs)
        extracted = 0
        src_fs_prefix = f"Odoo/filestore/{build_db}/"
        with tarfile.open(fileobj=tar_fs, mode="r") as tar:
            for member in tar.getmembers():
                if not member.name.startswith(src_fs_prefix) and member.name != f"Odoo/filestore/{build_db}":
                    continue
                rel = member.name[len(src_fs_prefix):]
                if not rel:
                    continue
                member.name = rel
                tar.extract(member, template_filestore_path)
                if not member.isdir():
                    extracted += 1
        logger.info("Filestore extracted to %s (%d files)", template_filestore_path, extracted)

    except docker.errors.NotFound:
        logger.info(
            "Odoo did not create data dir during init (normal for --stop-after-init). "
            "The template filestore at %s is empty; environments will start with an empty filestore.",
            template_filestore_path,
        )

    odoo_uid_gid = get_odoo_uid_gid(client, odoo_image)
    uid_str, gid_str = odoo_uid_gid.split(":")
    uid, gid = int(uid_str), int(gid_str)
    for root, dirs, files in os.walk(template_filestore_path):
        os.chown(root, uid, gid)
        for name in dirs + files:
            os.chown(os.path.join(root, name), uid, gid)

    temp_container.remove(v=True)
    logger.info("Temporary container removed")

    _exec_sql(client, settings, f"DROP DATABASE IF EXISTS {build_db} WITH (FORCE);")
    logger.info("Temporary database dropped")

    logger.info("Template generation complete, loading into template DB")
    result = reload_template(settings, template_name=template_name)

    metadata = {"odoo_image": odoo_image}
    metadata_path = settings.get_template_metadata_path(template_name)
    with open(metadata_path, "w") as f:
        json.dump(metadata, f, indent=2)
    logger.info("Template metadata saved to %s", metadata_path)

    result["generated_dump"] = template_sql_path
    result["generated_filestore"] = template_filestore_path
    result["filestore_files"] = sum(1 for _ in pathlib.Path(template_filestore_path).rglob("*") if _.is_file())
    return result


_TEMPLATE_EDITOR_CONTAINER = "flow-template-editor"
_TEMPLATE_EDITOR_BRANCH = "__template__"


def template_up(
    settings: Settings,
    odoo_image: str,
    template_name: str,
) -> dict[str, str]:
    client = get_client()
    tpl_db = get_template_db_name(template_name, settings.instance_id)

    try:
        existing = client.containers.get(_TEMPLATE_EDITOR_CONTAINER)
        if existing.status == "running":
            existing.reload()
            ports = existing.ports.get("8069/tcp")
            host_port = ports[0]["HostPort"] if ports else "?"
            url = f"http://{settings.external_host}:{host_port}"
            raise ConflictError(
                f"Template editor is already running at {url}. "
                f"Use --template-down to stop it first."
            )
        existing.remove(force=True)
    except docker.errors.NotFound:
        pass

    try:
        db_container = client.containers.get(settings.shared_db_container)
        if db_container.status != "running":
            raise PrerequisiteNotMetError(
                f"{settings.shared_db_container} is not running. Run init first."
            )
    except docker.errors.NotFound:
        raise PrerequisiteNotMetError(
            f"{settings.shared_db_container} not found. Run init first."
        )

    _wait_pg_ready(client, settings)

    if not _db_exists(client, settings, tpl_db):
        raise PrerequisiteNotMetError(
            f"Template database '{tpl_db}' not found. "
            f"Run init-template first."
        )

    _exec_sql(
        client,
        settings,
        f"UPDATE pg_database SET datistemplate=false WHERE datname='{tpl_db}';",
    )
    logger.info("Template flag removed from %s", tpl_db)

    template_filestore_path = settings.get_template_filestore_path(template_name)
    os.makedirs(template_filestore_path, exist_ok=True)

    odoo_uid_gid = get_odoo_uid_gid(client, odoo_image)
    uid_str, gid_str = odoo_uid_gid.split(":")
    uid, gid = int(uid_str), int(gid_str)
    for root, dirs, files in os.walk(template_filestore_path):
        os.chown(root, uid, gid)
        for name in dirs + files:
            os.chown(os.path.join(root, name), uid, gid)

    sessions_path = os.path.join(settings.get_template_dir(template_name), "sessions")
    os.makedirs(sessions_path, mode=0o777, exist_ok=True)
    os.chmod(sessions_path, 0o777)
    for root_dir, dirs, files in os.walk(sessions_path):
        os.chown(root_dir, uid, gid)
        for name in dirs + files:
            os.chown(os.path.join(root_dir, name), uid, gid)

    from oduflow.port_registry import allocate_port
    from oduflow.docker_ops.env_ops import _get_used_ports

    used_ports = _get_used_ports(client, settings)
    host_port = allocate_port(
        settings.port_registry_path,
        _TEMPLATE_EDITOR_BRANCH,
        settings.port_range_start,
        settings.port_range_end,
        used_ports=used_ports,
    )

    odoo_env = {
        "HOST": settings.shared_db_container,
        "USER": settings.db_user,
        "PASSWORD": settings.db_password,
    }
    odoo_volumes = {
        template_filestore_path: {
            "bind": f"/var/lib/odoo/.local/share/Odoo/filestore/{tpl_db}",
            "mode": "rw",
        },
        sessions_path: {
            "bind": "/var/lib/odoo/.local/share/Odoo/sessions",
            "mode": "rw",
        },
    }
    odoo_conf = _resolve_conf("odoo.conf")
    if odoo_conf.exists():
        odoo_volumes[str(odoo_conf)] = {"bind": "/etc/odoo/odoo.conf", "mode": "ro"}

    container = client.containers.run(
        odoo_image,
        name=_TEMPLATE_EDITOR_CONTAINER,
        detach=True,
        network=settings.shared_network,
        ports={"8069/tcp": host_port},
        environment=odoo_env,
        volumes=odoo_volumes,
        labels={settings.managed_label: "true"},
        command=f"odoo -d {tpl_db} --dev=xml",
    )

    metadata_path = settings.get_template_metadata_path(template_name)
    metadata = {}
    if os.path.isfile(metadata_path):
        with open(metadata_path) as f:
            metadata = json.load(f)
    metadata["odoo_image"] = odoo_image
    with open(metadata_path, "w") as f:
        json.dump(metadata, f, indent=2)

    url = f"http://{settings.external_host}:{host_port}"
    logger.info("Template editor started at %s (container=%s)", url, _TEMPLATE_EDITOR_CONTAINER)

    return {
        "status": "running",
        "url": url,
        "container": _TEMPLATE_EDITOR_CONTAINER,
        "database": tpl_db,
        "filestore": template_filestore_path,
    }


def template_down(settings: Settings, template_name: str) -> dict[str, str]:
    client = get_client()
    tpl_db = get_template_db_name(template_name, settings.instance_id)

    try:
        container = client.containers.get(_TEMPLATE_EDITOR_CONTAINER)
    except docker.errors.NotFound:
        raise NotFoundError(
            f"Template editor container '{_TEMPLATE_EDITOR_CONTAINER}' is not running."
        )

    if container.status == "running":
        container.stop()
    container.remove(v=True)
    logger.info("Template editor container removed")

    from oduflow.port_registry import release_port
    release_port(settings.port_registry_path, _TEMPLATE_EDITOR_BRANCH)

    _wait_pg_ready(client, settings)

    template_sql_path = settings.get_template_sql_path(template_name)
    os.makedirs(os.path.dirname(template_sql_path), exist_ok=True)

    db_container = client.containers.get(settings.shared_db_container)
    dump_file = f"/tmp/{tpl_db}.dump"
    dump_cmd = [
        "pg_dump", "-U", settings.db_user, "-Fc", "-f", dump_file, tpl_db,
    ]

    logger.info("Dumping template database %s", tpl_db)
    exit_code, output = db_container.exec_run(dump_cmd)
    if exit_code != 0:
        msg = output.decode("utf-8") if isinstance(output, bytes) else str(output)
        raise ExternalCommandError("pg_dump", exit_code, msg)

    chunks, _ = db_container.get_archive(dump_file)
    raw = b"".join(chunks)
    tar_stream = io.BytesIO(raw)
    with tarfile.open(fileobj=tar_stream, mode="r") as tar:
        member = tar.getmembers()[0]
        f = tar.extractfile(member)
        if f is None:
            raise ExternalCommandError("get_archive", -1, "Could not extract dump from tar")
        with open(template_sql_path, "wb") as out:
            out.write(f.read())

    logger.info("Dump saved to %s", template_sql_path)

    _exec_sql(
        client,
        settings,
        f"UPDATE pg_database SET datistemplate=true WHERE datname='{tpl_db}';",
    )
    logger.info("Template flag restored on %s", tpl_db)

    return {
        "status": "stopped",
        "dump": template_sql_path,
        "filestore": settings.get_template_filestore_path(template_name),
        "database": tpl_db,
    }


def publish_env_as_template(settings: Settings, branch_name: str, template_name: str) -> dict[str, str]:
    from oduflow.docker_ops import env_ops
    from oduflow.naming import get_db_name, get_filestore_paths

    client = get_client()
    tpl_db = get_template_db_name(template_name, settings.instance_id)
    env_db = get_db_name(branch_name, settings.instance_id)

    if not _db_exists(client, settings, env_db):
        raise NotFoundError(f"Database '{env_db}' for branch '{branch_name}' not found.")

    _wait_pg_ready(client, settings)
    db_container = client.containers.get(settings.shared_db_container)

    # 1. pg_dump branch DB → new template dump
    dump_path = settings.get_template_sql_path(template_name)
    os.makedirs(os.path.dirname(dump_path), exist_ok=True)
    dump_file = f"/tmp/{env_db}.dump"
    dump_cmd = ["pg_dump", "-U", settings.db_user, "-Fc", "-f", dump_file, env_db]

    logger.info("Dumping branch database %s", env_db)
    exit_code, output = db_container.exec_run(dump_cmd)
    if exit_code != 0:
        msg = output.decode("utf-8") if isinstance(output, bytes) else str(output)
        raise ExternalCommandError("pg_dump", exit_code, msg)

    chunks, _ = db_container.get_archive(dump_file)
    raw = b"".join(chunks)
    tar_stream = io.BytesIO(raw)
    with tarfile.open(fileobj=tar_stream, mode="r") as tar:
        member = tar.getmembers()[0]
        f = tar.extractfile(member)
        if f is None:
            raise ExternalCommandError("get_archive", -1, "Could not extract dump from tar")
        with open(dump_path, "wb") as out:
            out.write(f.read())

    logger.info("Branch dump saved to %s", dump_path)

    # 2. Reload template DB from new dump
    reload_template(settings, template_name=template_name, dump_path=dump_path)

    # 3. Collect active branches (excluding the promoted one is fine, it keeps working)
    active_envs = env_ops.list_environments(settings)
    active_branches = [e["branch"] for e in active_envs]

    # 4. Snapshot the promoted branch's merged filestore (while overlay is still mounted)
    branch_paths = get_filestore_paths(branch_name, settings.workspaces_dir)
    branch_merged = branch_paths["merged"]
    template_filestore_path = settings.get_template_filestore_path(template_name)

    if os.path.isdir(branch_merged) and os.path.ismount(branch_merged):
        # Overlay-mounted filestore — snapshot while still mounted
        snapshot_dir = branch_merged + "_snapshot"
        if os.path.exists(snapshot_dir):
            shutil.rmtree(snapshot_dir)
        shutil.copytree(branch_merged, snapshot_dir)
        logger.info("Snapshot of merged filestore created for branch %s", branch_name)
    elif os.path.isdir(branch_merged) and not os.path.ismount(branch_merged):
        # Plain (non-overlay) filestore — e.g. env created with template=none
        snapshot_dir = branch_merged + "_snapshot"
        if os.path.exists(snapshot_dir):
            shutil.rmtree(snapshot_dir)
        shutil.copytree(branch_merged, snapshot_dir)
        logger.info("Snapshot of plain filestore created for branch %s", branch_name)
    else:
        snapshot_dir = None
        logger.warning("Branch filestore %s not found, skipping filestore update", branch_merged)

    # 5. Unmount all overlays
    for branch in active_branches:
        try:
            env_ops._unmount_filestore(branch, settings)
            logger.info("Unmounted overlay for branch %s", branch)
        except Exception as e:
            logger.warning("Could not unmount overlay for %s: %s", branch, e)

    # 6. Replace template filestore with snapshot
    if snapshot_dir and os.path.isdir(snapshot_dir):
        if os.path.exists(template_filestore_path):
            shutil.rmtree(template_filestore_path)

        os.makedirs(os.path.dirname(template_filestore_path), exist_ok=True)
        try:
            os.rename(snapshot_dir, template_filestore_path)
        except OSError:
            shutil.copytree(snapshot_dir, template_filestore_path)
            shutil.rmtree(snapshot_dir)
        logger.info("Template filestore replaced from branch %s", branch_name)

        promoted_container_name = f"{settings.prefix}{branch_name.replace('/', '-')}-odoo"
        try:
            pc = client.containers.get(promoted_container_name)
            promoted_image = pc.image.tags[0] if pc.image.tags else "odoo:17.0"
        except (docker.errors.NotFound, IndexError):
            promoted_image = "odoo:17.0"
        odoo_uid_gid = get_odoo_uid_gid(client, promoted_image)
        uid_str, gid_str = odoo_uid_gid.split(":")
        uid, gid = int(uid_str), int(gid_str)
        for root, dirs, files in os.walk(template_filestore_path):
            os.chown(root, uid, gid)
            for name in dirs + files:
                os.chown(os.path.join(root, name), uid, gid)
        logger.info("Template filestore chowned to %s", odoo_uid_gid)

    # 7. Reset filestores to new template baseline
    for branch in active_branches:
        try:
            bp = get_filestore_paths(branch, settings.workspaces_dir)

            if os.path.ismount(bp["merged"]):
                env_ops._unmount_filestore(branch, settings)
                deadline = time.time() + 3.0
                while time.time() < deadline and os.path.ismount(bp["merged"]):
                    time.sleep(0.1)
                if os.path.ismount(bp["merged"]):
                    logger.warning(
                        "Overlay for %s still mounted after retry, skipping remount",
                        branch,
                    )
                    continue

            for key in ("upper", "work"):
                d = bp[key]
                if os.path.isdir(d):
                    shutil.rmtree(d)
                    os.makedirs(d, mode=0o777, exist_ok=True)

            odoo_container_name = f"{settings.prefix}{branch.replace('/', '-')}-odoo"
            try:
                container = client.containers.get(odoo_container_name)
                image = container.image.tags[0] if container.image.tags else "odoo:17.0"
            except (docker.errors.NotFound, IndexError):
                image = "odoo:17.0"

            env_db = get_db_name(branch, settings.instance_id)
            env_ops._mount_filestore(client, settings, branch, env_db, image, {}, template_name=template_name)
            logger.info("Filestore reset for branch %s", branch)

            try:
                container = client.containers.get(odoo_container_name)
                container.restart(timeout=10)
                logger.info("Restarted container %s", odoo_container_name)
            except docker.errors.NotFound:
                pass
        except Exception as e:
            logger.warning("Could not reset filestore for %s: %s", branch, e)

    # Save template metadata from promoted environment
    metadata_path = settings.get_template_metadata_path(template_name)
    promoted_container_name = f"{settings.prefix}{branch_name.replace('/', '-')}-odoo"
    metadata = {}
    try:
        pc = client.containers.get(promoted_container_name)
        metadata["odoo_image"] = pc.labels.get(settings.image_label, "")
        metadata["repo_url"] = pc.labels.get(settings.repo_label, "")
        raw_extras = pc.labels.get("oduflow.extra_addons", "")
        if raw_extras:
            try:
                parsed = json.loads(raw_extras)
                metadata["extra_addons"] = _normalize_extra_addons(parsed, settings.default_branch)
            except (json.JSONDecodeError, TypeError):
                pass
    except docker.errors.NotFound:
        pass
    with open(metadata_path, "w") as f:
        json.dump(metadata, f, indent=2)
    logger.info("Template metadata saved to %s", metadata_path)

    return {
        "status": "promoted",
        "branch": branch_name,
        "dump": dump_path,
        "filestore": template_filestore_path,
        "template_db": tpl_db,
    }


def destroy_system(settings: Settings) -> dict[str, str]:
    client = get_client()
    logger.info("Destroying system")

    filters = {"label": [f"{settings.managed_label}=true"]}
    containers = client.containers.list(all=True, filters=filters)
    system_names = {settings.shared_db_container, settings.traefik_container}
    env_containers = [
        c for c in containers
        if c.labels.get(settings.branch_label) and c.name not in system_names
    ]
    svc_containers = [
        c for c in containers
        if c.labels.get("oduflow.service") and c.name not in system_names
    ]
    blocking = env_containers + svc_containers
    if blocking:
        names = [c.name for c in blocking]
        from oduflow.errors import ConflictError
        if svc_containers and not env_containers:
            raise ConflictError(
                f"Active environments/services exist: {', '.join(names)}. Delete them first."
            )
        elif env_containers and not svc_containers:
            raise ConflictError(
                f"Active environments exist: {', '.join(names)}. Delete them first."
            )
        else:
            raise ConflictError(
                f"Active environments/services exist: {', '.join(names)}. Delete them first."
            )

    removed: list[str] = []

    _destroy_traefik(client, settings, removed)

    try:
        db = client.containers.get(settings.shared_db_container)
        db.stop()
        db.remove(v=True)
        removed.append(settings.shared_db_container)
    except docker.errors.NotFound:
        pass

    try:
        vol = client.volumes.get(settings.shared_db_volume)
        vol.remove()
        removed.append(settings.shared_db_volume)
    except docker.errors.NotFound:
        pass

    try:
        net = client.networks.get(settings.shared_network)
        net.remove()
        removed.append(settings.shared_network)
    except docker.errors.NotFound:
        pass

    logger.info("System destroyed, removed=%s", removed)
    return {"status": "destroyed", "removed": ", ".join(removed)}


def drop_template(settings: Settings, template_name: str) -> dict[str, str]:
    client = get_client()
    tpl_db = get_template_db_name(template_name, settings.instance_id)

    if _db_exists(client, settings, tpl_db):
        _wait_pg_ready(client, settings)
        _exec_sql(client, settings, f"UPDATE pg_database SET datistemplate=false WHERE datname='{tpl_db}';")
        _exec_sql(client, settings, f'DROP DATABASE IF EXISTS "{tpl_db}" WITH (FORCE);')
        logger.info("Dropped template DB %s", tpl_db)

    template_dir_path = settings.get_template_dir(template_name)
    if os.path.isdir(template_dir_path):
        shutil.rmtree(template_dir_path)
        logger.info("Removed template directory %s", template_dir_path)

    return {"status": "dropped", "template_name": template_name, "template_db": tpl_db}


def list_templates(settings: Settings) -> list[dict]:
    client = get_client()
    templates = settings.list_templates()
    result = []
    for template_name in templates:
        tpl_db = get_template_db_name(template_name, settings.instance_id)
        has_sql = os.path.isfile(settings.get_template_sql_path(template_name))
        has_filestore = os.path.isdir(settings.get_template_filestore_path(template_name))
        db_loaded = _db_exists(client, settings, tpl_db)
        metadata = {}
        metadata_path = settings.get_template_metadata_path(template_name)
        if os.path.isfile(metadata_path):
            with open(metadata_path) as f:
                metadata = json.load(f)
        result.append({
            "template_name": template_name,
            "template_db": tpl_db,
            "has_sql": has_sql,
            "has_filestore": has_filestore,
            "db_loaded": db_loaded,
            "odoo_image": metadata.get("odoo_image", ""),
            "repo_url": metadata.get("repo_url", ""),
            "extra_addons": _normalize_extra_addons(
                metadata.get("extra_addons", {}),
                settings.default_branch,
            ),
        })
    return result



"""Shared helpers for calling MCP tools and CLI commands directly."""

from __future__ import annotations

import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parent.parent / "src"))

from oduflow.server import mcp  # noqa: E402


def call_tool(name: str, **kwargs):
    """Look up a registered MCP tool by name and invoke it with *kwargs*."""
    tool = mcp._tool_manager._tools[name]
    return tool.fn(**kwargs)


def list_tools() -> list[str]:
    """Return sorted list of all registered tool names."""
    return sorted(mcp._tool_manager._tools.keys())


def call_cli(command: str, **kwargs) -> str:
    """Run a CLI command (init / destroy) and return formatted output."""
    from oduflow.docker_ops import system_ops
    from oduflow.settings import Settings

    settings = Settings.from_env()

    if command == "init":
        result = system_ops.init_system(
            settings,
            version=kwargs.get("version", "15.0"),
            force=kwargs.get("force", False),
        )
        return f"System {result['status']}."

    if command == "destroy":
        result = system_ops.destroy_system(settings)
        return (
            f"System {result['status']}.\n"
            f"Removed: {result['removed']}"
        )

    raise ValueError(f"Unknown CLI command: {command}")



import json
import os
import logging

logger = logging.getLogger("oduflow")


def _load_registry(registry_path: str) -> dict[str, int]:
    """Load port registry from disk. Returns empty dict if file doesn't exist or is corrupt."""
    if not os.path.isfile(registry_path):
        return {}
    try:
        with open(registry_path) as f:
            data = json.load(f)
        return {k: int(v) for k, v in data.items()}
    except (json.JSONDecodeError, ValueError, OSError) as e:
        logger.warning("Could not load port registry %s: %s", registry_path, e)
        return {}


def _save_registry(registry_path: str, registry: dict[str, int]) -> None:
    """Atomically save port registry to disk."""
    tmp_path = registry_path + ".tmp"
    with open(tmp_path, "w") as f:
        json.dump(registry, f, indent=2)
        f.flush()
        os.fsync(f.fileno())
    os.replace(tmp_path, registry_path)


def allocate_port(
    registry_path: str,
    branch_name: str,
    port_range_start: int,
    port_range_end: int,
    used_ports: set[int] | None = None,
) -> int:
    """Allocate a stable port for a branch.

    If the branch already has a port in the registry AND it's not used by another
    container (checked via used_ports), reuse it. Otherwise allocate the next free port.

    Args:
        registry_path: Path to ports.json
        branch_name: Branch to allocate port for
        port_range_start: Start of port range (inclusive)
        port_range_end: End of port range (exclusive)
        used_ports: Set of ports currently in use by OTHER branches' Docker containers.
                    If None, no conflict checking against Docker is done.

    Returns:
        The allocated port number.

    Raises:
        FlowError if no free ports available.
    """
    from oduflow.errors import FlowError

    if used_ports is None:
        used_ports = set()

    registry = _load_registry(registry_path)

    if branch_name in registry:
        existing_port = registry[branch_name]
        if port_range_start <= existing_port < port_range_end and existing_port not in used_ports:
            return existing_port

    occupied = set(registry.values()) | used_ports

    for port in range(port_range_start, port_range_end):
        if port not in occupied:
            registry[branch_name] = port
            _save_registry(registry_path, registry)
            logger.info("Allocated port %d for branch '%s'", port, branch_name)
            return port

    raise FlowError(
        f"No free ports in range {port_range_start}-{port_range_end}. "
        f"Delete unused environments to free ports."
    )


def release_port(registry_path: str, branch_name: str) -> None:
    """Remove port assignment for a branch."""
    registry = _load_registry(registry_path)
    if branch_name in registry:
        port = registry.pop(branch_name)
        _save_registry(registry_path, registry)
        logger.info("Released port %d for branch '%s'", port, branch_name)


def get_port(registry_path: str, branch_name: str) -> int | None:
    """Get the assigned port for a branch, or None if not assigned."""
    registry = _load_registry(registry_path)
    return registry.get(branch_name)
